from eprllib.Agents.Triggers.BaseTrigger import BaseTrigger
from eprllib.Agents.Triggers.TriggerSpec import TriggerSpec
import pytest

class TestTriggerspec:

    def test___getitem___access_valid_attribute(self):
        """
        Test that __getitem__ method correctly returns the value of a valid attribute.
        """
        class DummyTrigger(BaseTrigger):
            pass

        trigger_spec = TriggerSpec(trigger_fn=DummyTrigger, trigger_fn_config={'key': 'value'})

        assert trigger_spec['trigger_fn'] == DummyTrigger
        assert trigger_spec['trigger_fn_config'] == {'key': 'value'}

    def test___getitem___non_existent_attribute(self):
        """
        Test the __getitem__ method with a non-existent attribute.
        This should raise an AttributeError as the method directly uses getattr().
        """
        trigger_spec = TriggerSpec()
        with pytest.raises(AttributeError):
            trigger_spec['non_existent_key']

    # def test___init___default_values(self):
    #     """
    #     Test the __init__ method of TriggerSpec with default values.

    #     This test verifies that the TriggerSpec can be instantiated with default values
    #     and that the attributes are set correctly.
    #     """
    #     trigger_spec = TriggerSpec()
    #     assert trigger_spec.trigger_fn == NotImplemented
    #     assert trigger_spec.trigger_fn_config == {}

    # def test___init___invalid_trigger_fn(self):
    #     """
    #     Test that initializing TriggerSpec with a trigger_fn that is not a subclass of BaseTrigger
    #     raises a ValueError.
    #     """
    #     with pytest.raises(ValueError) as exc_info:
    #         TriggerSpec(trigger_fn=str).build()
    #     assert "The trigger function must be based on BaseTrigger class" in str(exc_info.value)

    # def test___init___invalid_trigger_fn_config(self):
    #     """
    #     Test that initializing TriggerSpec with a trigger_fn_config that is not a dictionary
    #     raises a ValueError.
    #     """
    #     class DummyTrigger(BaseTrigger):
    #         pass

    #     with pytest.raises(ValueError) as exc_info:
    #         TriggerSpec(trigger_fn=DummyTrigger, trigger_fn_config=[]).build()
    #     assert "The configuration for the trigger function must be a dictionary" in str(exc_info.value)

    # def test___init___not_implemented_trigger_fn(self):
    #     """
    #     Test that initializing TriggerSpec with NotImplemented as trigger_fn
    #     raises a NotImplementedError when build() is called.
    #     """
    #     trigger_spec = TriggerSpec(trigger_fn=NotImplemented)
    #     with pytest.raises(ValueError) as exc_info:
    #         trigger_spec.build()
    #     assert "The trigger function must be defined" in str(exc_info.value)

    def test___setitem___invalid_key(self):
        """
        Test that __setitem__ raises a KeyError when an invalid key is provided.
        """
        trigger_spec = TriggerSpec()
        with pytest.raises(KeyError) as excinfo:
            trigger_spec['invalid_key'] = 'some_value'
        assert str(excinfo.value) == "'Invalid key: invalid_key.'"

    def test___setitem___invalid_key_2(self):
        """
        Test that __setitem__ raises a KeyError when an invalid key is provided.

        This test verifies that when attempting to set an item with a key that
        is not in the valid keys of the TriggerSpec instance, a KeyError is raised
        with an appropriate error message.
        """
        trigger_spec = TriggerSpec()
        invalid_key = "invalid_key"

        with pytest.raises(KeyError) as excinfo:
            trigger_spec[invalid_key] = "some_value"

        assert str(excinfo.value) == f"'Invalid key: {invalid_key}.'"

    def test___setitem___valid_key(self):
        """
        Test that __setitem__ successfully sets a value for a valid key.
        This test covers the path where the key is in valid_keys.
        """
        trigger_spec = TriggerSpec(trigger_fn=BaseTrigger, trigger_fn_config={})

        # Set a value for an existing attribute
        trigger_spec['trigger_fn'] = None

        assert trigger_spec.trigger_fn is None

    # def test_build_2(self):
    #     """
    #     Test the build method when trigger_fn is not NotImplemented,
    #     but is not a subclass of BaseTrigger, and trigger_fn_config is not a dict.
    #     Expects a ValueError to be raised.
    #     """
    #     # Create a TriggerSpec instance with invalid trigger_fn and trigger_fn_config
    #     trigger_spec = TriggerSpec(trigger_fn=str, trigger_fn_config="invalid_config")

    #     # Assert that ValueError is raised when calling build()
    #     with pytest.raises(ValueError) as excinfo:
    #         trigger_spec.build()

    #     # Check that the error message indicates the trigger_fn is not based on BaseTrigger
    #     assert "The trigger function must be based on BaseTrigger class" in str(excinfo.value)

    # def test_build_3_invalid_trigger_fn_config(self):
    #     """
    #     Test that build() raises a ValueError when trigger_fn is NotImplemented
    #     and trigger_fn_config is not a dictionary.
    #     """
    #     trigger_spec = TriggerSpec(trigger_fn=NotImplemented, trigger_fn_config="invalid_config")

    #     with pytest.raises(ValueError) as context:
    #         trigger_spec.build()

    #     assert str(context.value) == "The trigger function must be defined."

    # def test_build_4(self):
    #     """
    #     Test the build method when trigger_fn is NotImplemented.

    #     This test verifies that the build method raises a ValueError
    #     when the trigger_fn is set to NotImplemented, which violates
    #     the requirement that trigger_fn must be defined.
    #     """
    #     trigger_spec = TriggerSpec(trigger_fn=NotImplemented, trigger_fn_config={})

    #     with pytest.raises(ValueError) as exc_info:
    #         trigger_spec.build()

    #     assert str(exc_info.value) == "The trigger function must be defined."

    # def test_build_invalid_trigger_fn(self):
    #     """
    #     Test the build method when trigger_fn is NotImplemented.

    #     This test verifies that the build method raises a ValueError
    #     when the trigger_fn is set to NotImplemented.
    #     """
    #     trigger_spec = TriggerSpec(trigger_fn=NotImplemented, trigger_fn_config={})

    #     with pytest.raises(ValueError) as context:
    #         trigger_spec.build()

    #     assert str(context.value) == "The trigger function must be defined."

    # def test_build_with_non_basetrigger_function(self):
    #     """
    #     Test the build method when the trigger function is not a subclass of BaseTrigger.
    #     This should raise a ValueError.
    #     """
    #     class NonBaseTrigger:
    #         pass

    #     trigger_spec = TriggerSpec(trigger_fn=NonBaseTrigger)
    #     with pytest.raises(ValueError) as context:
    #         trigger_spec.build()
        
    #     assert str(context.value) == f"The trigger function must be based on BaseTrigger class but {type(NonBaseTrigger)} was given."

    # def test_build_with_non_dict_config(self):
    #     """
    #     Test the build method when the trigger function config is not a dictionary.
    #     This should raise a ValueError.
    #     """
    #     class MockTrigger(BaseTrigger):
    #         pass

    #     trigger_spec = TriggerSpec(trigger_fn=MockTrigger, trigger_fn_config="not a dict")
    #     with pytest.raises(ValueError) as context:
    #         trigger_spec.build()
    #     assert str(context.value) == f"The configuration for the trigger function must be a dictionary but {type(trigger_spec.trigger_fn_config)} was given."

    def test_build_with_undefined_trigger_function(self):
        """
        Test the build method when the trigger function is not defined (NotImplemented).
        This should raise a ValueError.
        """
        trigger_spec = TriggerSpec()
        with pytest.raises(ValueError) as context:
            trigger_spec.build()
        assert str(context.value) == "The trigger function must be defined."



"""
Default Filter
===============

This module contains the default filter class for preprocessing observations before they are fed to the agent.
The `DefaultFilter` class extends the `BaseFilter` class and provides a basic implementation that can be used
as-is or extended to create custom filters.
"""
import numpy as np
from typing import Any, Dict
from numpy.typing import NDArray
from eprllib.Agents.Filters.BaseFilter import BaseFilter
from eprllib.Utils.annotations import override
from eprllib import logger

class DefaultFilter(BaseFilter):
    """
    Default filter class for preprocessing observations.

    This class extends the `BaseFilter` class and provides a basic implementation that can be used
    as-is or extended to create custom filters. The `get_filtered_obs` method returns the agent
    states as a numpy array of float32 values.
    """
    def __init__(
        self,
        filter_fn_config: Dict[str, Any]
    ):
        """
        Initializes the DefaultFilter class.

        Args:
            filter_fn_config (Dict[str, Any]): Configuration dictionary for the filter function.
            This configuration can include settings that affect how the observations are filtered.
        """
        super().__init__(filter_fn_config)
        
    @override(BaseFilter)
    def get_filtered_obs(
        self,
        env_config: Dict[str, Any],
        agent_states: Dict[str, Any],
    ) -> NDArray[np.float32]:
        """
        Returns the filtered observations for the agent based on the environment configuration
        and agent states. This method processes the raw observations according to the filter
        configuration specified in filter_fn_config.

        Args:
            env_config (Dict[str, Any]): Configuration dictionary for the environment. Can include 
            settings that affect how the observations are filtered.
            
            agent_states (Dict[str, Any], optional): Dictionary containing the states of the agent.

        Returns:
            NDarray: Filtered observations as a numpy array of float32 values.
            
        Raises:
            TypeError: If agent_states is not a dictionary.
            ValueError: If agent_states is empty or contains non-numeric values.
        """
        # Check if the agent_states dictionary is empty
        if not agent_states:
            msg = "agent_states dictionary is empty"
            logger.error(msg)
            raise ValueError(msg)
        
        # Check if all values in the agent_states dictionary are numeric
        if not all(isinstance(value, (int, float)) for value in agent_states.values()):
            msg = "All values in agent_states must be numeric"
            logger.error(msg)
            raise ValueError(msg)
        
        return np.array(list(agent_states.values()), dtype='float32')
    


"""
Fully Shared Parameters Agents Connector
=========================================

This module implements a fully shared parameters policy for the observation function.

The flat observation for each agent must be a numpy array and the size must be equal for all agents. To 
achieve this, a concatenation of variables is performed, which includes:

- AgentID one-hot encoded vector
- Agent state
- Actuators state (including other agents)

Notes:
------
- The number of actuators for other agents is variable. Consider changing the actuator state for agent action or 
  finding a way to expand the observation space to adapt to different observation sizes.
"""
import time
import numpy as np
from gymnasium.spaces import Box
from typing import Any, Dict, Tuple, Optional
from eprllib.AgentsConnectors.BaseConnector import BaseConnector
from eprllib.Utils.annotations import override
import eprllib.Utils.observation_utils as observation_utils
from eprllib.Utils.connector_utils import (
    set_variables_in_obs,
    set_internal_variables_in_obs,
    set_meters_in_obs,
    set_simulation_parameters_in_obs,
    set_zone_simulation_parameters_in_obs,
    set_prediction_variables_in_obs,
    set_other_obs_in_obs,
    set_user_occupation_forecast_in_obs
    )
from eprllib import logger

class FullySharedParametersConnector(BaseConnector):
    def __init__(
        self,
        connector_fn_config: Dict[str, Any]
    ):
        """
        This class implements a fully shared parameters policy for the observation function.

        Args:
            connector_fn_config (Dict[str, Any]): The configuration dictionary for the observation function.
            This must contain the key 'number_of_agents_total', which represents the maximum
            quantity for which the policy is prepared. It is related to the unitary vector.
        """
        super().__init__(connector_fn_config)
        self.number_of_agents_total: int = connector_fn_config['number_of_agents_total']
        self.number_of_actuators_total: int = connector_fn_config['number_of_actuators_total']
        self.agent_ids: Optional[Dict[str, Optional[int]]] = None
        self.actuator_ids: Optional[Dict[str, Optional[int]]] = None
    
    @override(BaseConnector)
    def get_agent_obs_dim(
        self,
        env_config: Dict[str,Any],
        agent: str
        ) -> Box:
        """
        Construct the observation space of the environment.

        Args:
            env_config (Dict[str, Any]): The environment configuration dictionary.
            agent (str, optional): The agent identifier.

        Returns:
            gym.Space: The observation space of the environment.
        """
        agent_list = [key for key in env_config["agents_config"].keys()]
        obs_space_len: int = 0
        self.obs_indexed[agent] = {}
        
        if len(agent_list) > self.number_of_agents_total:
            raise ValueError("The number of agents must be greater than the number of agents in the environment configuration.")
        
        if self.number_of_agents_total > 1:
            agents_used = self.number_of_agents_total
            for agent_name in agent_list:
                self.obs_indexed[agent][agent_name] = obs_space_len
                obs_space_len += 1
                agents_used -= 1
            # Add the rest of the agents as zero vectors
            if agents_used > 0:
                for i in range(agents_used):
                    self.obs_indexed[agent][f"not_used_agent_{i+1}"] = obs_space_len
                    obs_space_len += 1
        assert obs_space_len == len(self.obs_indexed[agent]), f"The observation space length must be equal to the number of indexed observations. Obs indexed:{len(self.obs_indexed[agent])} != Obs space len:{obs_space_len}."
        self.obs_indexed[agent], obs_space_len = set_variables_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_internal_variables_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_meters_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_simulation_parameters_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_zone_simulation_parameters_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_prediction_variables_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_other_obs_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_user_occupation_forecast_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        
        # Add the actuator state to all the agents obs_space_len
        # Check if at least one agent request the use_actuator_state
        use_actuator_state_flag = False
        for agent_name in agent_list:
            if env_config["agents_config"][agent_name]["observation"]['use_actuator_state']:
                use_actuator_state_flag = True
                break
        if use_actuator_state_flag:
            # Save the total amount of actuators in the environment to create a vector in set_agent_obs
            actuator_used = self.number_of_actuators_total
            for agent_name in agent_list:
                for actuator in range(len(env_config["agents_config"][agent_name]["action"]['actuators'])):
                    actuator_component_type = env_config["agents_config"][agent_name]["action"]['actuators'][actuator][0]
                    actuator_control_type = env_config["agents_config"][agent_name]["action"]['actuators'][actuator][1]
                    actuator_key = env_config["agents_config"][agent_name]["action"]['actuators'][actuator][2]
                    self.obs_indexed[agent][observation_utils.get_actuator_name(
                        agent,
                        actuator_component_type,
                        actuator_control_type,
                        actuator_key
                        )] = obs_space_len
                    obs_space_len += 1
                    actuator_used -= 1
            # Add the rest of the actuators as zero vectors
            if actuator_used > 0:
                for i in range(actuator_used):
                    self.obs_indexed[agent][f"not_used_actuator_{i+1}"] = obs_space_len
                    obs_space_len += 1
                    
            # chack that actuators is equal or minor to self.number_of_actuators_total
            if actuator_used < 0:
                logger.error("The total amount of actuators in the environment is greater than the number of actuators in the environment configuration.")
        
        assert obs_space_len > 0, "The observation space length must be greater than 0."
        assert len(self.obs_indexed[agent]) == obs_space_len, f"The observation space length must be equal to the number of indexed observations. Obs indexed:{len(self.obs_indexed)} != Obs space len:{obs_space_len}. The agent {agent} has the following indexed observations: {self.obs_indexed[agent]}."
        # obs_space_len += 1
        logger.debug(f"Observation space length for agent {agent}: {obs_space_len}")
        
        return Box(float("-inf"), float("inf"), (obs_space_len, ))
    
    @override(BaseConnector)
    def get_agent_obs_indexed(
        self,
        env_config: Dict[str, Any],
        agent: str
    ) -> Dict[str, int]:
        """
        Get a dictionary of the agent observation parameters and their respective index in the observation array.

        :param env_config: Environment configuration.
        :type env_config: Dict[str, Any]
        :param agent: Agent identifier, optional.
        :type agent: str, optional
        :return: Agent observation spaces.
        :rtype: gym.spaces.Space
        """
        if self.obs_indexed == {}:
            self.get_agent_obs_dim(env_config, agent)
        return self.obs_indexed[agent]
    
    @override(BaseConnector)
    def set_top_level_obs(
        self,
        env_config: Dict[str, Any],
        agent_states: Dict[str,Dict[str,Any]],
        dict_agents_obs: Dict[str,Any],
        infos: Dict[str, Dict[str, Any]],
        is_last_timestep: bool = False
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]:
        """This method construct the observation of the environment. 
        The observation is a concatenation of the following variables:
        - agent indicator
        - agent state
        - actuator state
        - internal variables
        - meters
        - simulation parameters
        - zone simulation parameters
        - weather prediction
        - other observation

        Args:
            env_config (Dict[str, Any]): Dict of an environment config builded.
            agent_states (Dict[str,Dict[str,Any]]): The agent actual states before filter.
            dict_agents_obs (Dict[str,Any]): The agent actual states after filter with ``ObservationFunction``.
            infos (Dict[str, Dict[str, Any]]): Agent infos dict.
            is_last_timestep (bool, optional): Flag to indicate the last timestep before finish the RunPeriod. Defaults to False.

        Raises:
            ValueError: _description_
            ValueError: _description_

        Returns:
            Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]: Returns a Tuple with the agents observations addapted to
            fully shared parameter policy, the infos dict updated, and a flag to tell that is the lowest-level agent (that means
            that the actions obtained after apply this observation into the policy shall be implemened in an actuator after 
            trigger the policy action in the ActionFunction of the agent.)
        """
        # Check that the lenght of all the dict_agents_obs key values are the same:
        if len(set([len(value) for value in dict_agents_obs.values()])) != 1:
            msg = "The lenght of all the dict_agents_obs key values must be the same when you use homogeneous fully shared parameters policy."
            logger.error(msg)
            raise ValueError(msg)
        
        # Add the ID vectors if it's needed
        if self.agent_ids is None:
            id = 0
            self.agent_ids = {agent: None for agent in env_config['agents_config'].keys()}
            for agent in env_config['agents_config'].keys():
                self.agent_ids.update({agent: id})
                id += 1
                if len(self.agent_ids) > self.number_of_agents_total:
                    msg = f"The agents found were: {env_config['agents_config'].keys()} with a total of {len(env_config['agents_config'].keys())}, that are greather than {self.number_of_agents_total}."
                    logger.error(msg)
                    raise ValueError(msg)

            if len(self.agent_ids) > self.number_of_agents_total:
                msg = f"The agents found were: {env_config['agents_config'].keys()} with a total of {len(env_config['agents_config'].keys())}, that are greather than {self.number_of_agents_total}."
                logger.error(msg)
                raise ValueError(msg)
        
        if self.actuator_ids is None:
            id = 0
            self.actuator_ids = {}
            for agent in env_config['agents_config'].keys():
                for actuator_config in env_config["agents_config"][agent]["action"]["actuators"]:
                    self.actuator_ids.update({observation_utils.get_actuator_name(agent,actuator_config[0],actuator_config[1],actuator_config[2]): id})
                    id += 1
                    if len(self.actuator_ids) > self.number_of_actuators_total:
                        msg = f"The actuators found were: {self.actuator_ids.keys()} with a total of {len(self.actuator_ids.keys())}, that are greather than {self.number_of_actuators_total}."
                        logger.error(msg)
                        raise ValueError(msg)
        
        # agents in this timestep
        agent_list = [key for key in dict_agents_obs.keys()]
        # Add agent indicator for the observation for each agent
        actuator_names: Dict[str, Dict[str, Any]] = {agent: {} for agent in agent_list}
        
        for agent in agent_list:
            # Remove from agent_states and save the actuator items.
            for actuator_config in env_config["agents_config"][agent]["action"]["actuators"]:
                actuator_name = observation_utils.get_actuator_name(agent,actuator_config[0],actuator_config[1],actuator_config[2])
                actuator_names[agent].update({actuator_name: agent_states[agent].get(actuator_name, -2)})
                if actuator_names[agent][actuator_name] == -2:
                    logger.info(f"Looking for actuator: {actuator_name}")
                    logger.info(f"Available keys in agent_states[{agent}]: {agent_states[agent].keys()}")
                    time.sleep(10)
                    
            
            # Agent properties
            agent_id_vector = None
            
            # === AgentID one-hot enconde vector ===
            # If apply, add the igent ID vector for each agent obs
            if self.number_of_agents_total > 1:
                agent_id_vector = np.array([0]*self.number_of_agents_total)
                agent_id_vector[self.agent_ids[agent]] = 1
                
            
            if agent_id_vector is not None:
                dict_agents_obs[agent] = np.concatenate(
                    (
                        agent_id_vector,
                        dict_agents_obs[agent]
                    ),
                    dtype='float32'
                )
            
        # if apply, add the actuator state as a vector of all agents.
        use_actuator_state_flag = False
        for agent in agent_list:
            if env_config["agents_config"][agent]["observation"]['use_actuator_state']:
                use_actuator_state_flag = True
                
        if use_actuator_state_flag:
            
            actuator_id_vector = np.array([-2]*self.number_of_actuators_total)
            
            for agent in agent_list:
                for actuator in actuator_names[agent]:
                    actuator_id_vector[self.actuator_ids[actuator]] = actuator_names[agent][actuator]
            
            for agent in agent_list:        
                dict_agents_obs[agent] = np.concatenate(
                    (
                        dict_agents_obs[agent],
                        actuator_id_vector,
                    ),
                    dtype='float32'
                )
            
        return dict_agents_obs, infos, True



"""
Default Episode
================

This module contains the default implementation of the episode functions for the EnergyPlus environment.
"""

from typing import Dict, Any, List
from eprllib.Episodes.BaseEpisode import BaseEpisode
from eprllib.Utils.annotations import override

class DefaultEpisode(BaseEpisode):
    """
    This class provides the default implementation of the episode functions for the EnergyPlus environment.
    It inherits from the BaseEpisode class.
    """
    def __init__(
        self,
        episode_fn_config: Dict[str, Any] = {}
    ):
        """
        Initializes the DefaultEpisode object.

        Args:
            episode_fn_config (Dict[str, Any]): Configuration dictionary for the episode function.
        """
        super().__init__(episode_fn_config)

    @override(BaseEpisode)
    def get_episode_config(self, env_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Returns the episode configuration for the EnergyPlus environment.

        Args:
            env_config (Dict[str, Any]): The environment configuration.

        Returns:
            Dict[str, Any]: The episode configuration.
        """
        return super().get_episode_config(env_config)
    
    @override(BaseEpisode)
    def get_episode_agents(self, env_config: Dict[str, Any], possible_agents: List[str]) -> List[str]:
        """
        Returns the agents for the episode configuration in the EnergyPlus environment.

        Args:
            env_config (Dict[str, Any]): The environment configuration.
            possible_agents (List[str]): List of possible agents.

        Returns:
            Dict[str, Any]: The agents that are acting for the episode configuration. Default: possible_agents list.
        """
        return super().get_episode_agents(env_config, possible_agents)
    
    @override(BaseEpisode)
    def get_timestep_agents(self, env_config: Dict[str, Any], possible_agents: List[str]) -> List[str]:
        """
        Returns the agents for the timestep configuration in the EnergyPlus environment.

        Args:
            env_config (Dict[str, Any]): The environment configuration.
            possible_agents (List[str]): List of possible agents.

        Returns:
            Dict[str, Any]: The agents that are acting for the timestep configuration. Default: possible_agents list.
        """
        return super().get_timestep_agents(env_config, possible_agents)


"""
Example with Central Agent to control a mix-ventilated building
================================================================

"""
import os
os.environ['RAY_DEDUP_LOGS'] = '0'

import time
import json
from tempfile import TemporaryDirectory
import ray
from ray import air, tune
from ray.tune import register_env
from ray.rllib.algorithms.ppo.ppo import PPOConfig
from ray.rllib.policy.policy import PolicySpec

from eprllib.Environment.Environment import Environment
from eprllib.Environment.EnvironmentConfig import EnvironmentConfig
from eprllib.AgentsConnectors.DefaultConnector import DefaultConnector
from eprllib.Agents.AgentSpec import (
    AgentSpec,
    ObservationSpec,
    RewardSpec,
    ActionSpec,
    TriggerSpec,
    FilterSpec
)

from eprllib.Agents.Filters.DefaultFilter import DefaultFilter

from eprllib.examples.example_central_agent_files.trigger import CentralAgentTrigger
from eprllib.examples.example_central_agent_files.reward_function import IAQThermalComfortEnergyReward
from eprllib.examples.example_central_agent_files.episode_fn import task_cofiguration
from eprllib.examples.example_central_agent_files.policy_mapping import policy_map_fn

# read the json config file as a dict.
with open("src/eprllib/examples/example_central_agent_files/episode_fn_config.json", "r") as f:
    episode_config = json.load(f)

experiment_name = "example_central_agent"
name = "dnn_simple2"
tuning = False
restore = False
checkpoint_path = ""

eprllib_config = EnvironmentConfig()
eprllib_config.generals(
    epjson_path = "src/eprllib/examples/example_central_agent_files/model.epJSON",
    epw_path = "src/eprllib/examples/example_central_agent_files/weathers/ESP_PV_San.Sebastian-Igueldo.080270_TMYx.2004-2018.epw",
    output_path = TemporaryDirectory("output","",'C:/Users/grhen/Documents/Resultados_RLforEP').name,
    ep_terminal_output = False,
    timeout = 10,
    evaluation = False,
)
eprllib_config.connector(
    connector_fn = DefaultConnector,
    connector_fn_config = {},
)
eprllib_config.agents(
    agents_config = {
        "Central Agent": AgentSpec(
            observation = ObservationSpec(
                variables = [
                    ("Site Outdoor Air Drybulb Temperature", "Environment"),
                    ("Site Wind Speed", "Environment"),
                    ("Site Outdoor Air Relative Humidity", "Environment"),
                    ("Zone Mean Air Temperature", "Thermal Zone"),
                    ("Zone Air Relative Humidity", "Thermal Zone"),
                    ("Zone People Occupant Count", "Thermal Zone"),
                    ("Zone Air CO2 Concentration", "Thermal Zone"),
                    ("Fan Electricity Energy", "ExhaustFan"),
                ],
                simulation_parameters = {
                    'today_weather_horizontal_ir_at_time': True,
                },
                meters = [
                    "Electricity:Building",
                    "Cooling:DistrictCooling",
                    "Heating:DistrictHeatingWater"
                ],
                use_actuator_state = True,
                use_one_day_weather_prediction = True,
                prediction_hours = 3,
                prediction_variables = {
                    'outdoor_dry_bulb': True,
                },
                history_len = 6,
            ),
            action = ActionSpec(
                actuators = [
                    ("AirFlow Network Window/Door Opening", "Venting Opening Factor", "window_east"),
                    ("Schedule:Constant", "Schedule Value", "ventilation_factor"),
                    ("Schedule:Constant", "Schedule Value", "HVAC_OnOff"),
                ],
            ),
            filter= FilterSpec(
                filter_fn = DefaultFilter,
                filter_fn_config = {},
            ),
            trigger= TriggerSpec(
                trigger_fn = CentralAgentTrigger,
                trigger_fn_config = {
                    'window_actuator': ("AirFlow Network Window/Door Opening", "Venting Opening Factor", "window_east"),
                    'exhaust_fan_actuator': ("Schedule:Constant", "Schedule Value", "ventilation_factor"),
                    'hvac_availability_actuator': ("Schedule:Constant", "Schedule Value", "HVAC_OnOff"),
                },
            ),
            reward = RewardSpec(
                reward_fn = IAQThermalComfortEnergyReward,
                reward_fn_config = {
                    "thermal_zone": "Thermal Zone",
                    "co2_threshold": 800.0,
                    "exhaust_fan_name": "ExhaustFan",
                    'cooling_name': "Cooling:DistrictCooling",
                    'heating_name': "Heating:DistrictHeatingWater",
                    'cooling_energy_ref': 7100*(10*60), # From W to J in a timestep ot 10 minutes: P*t
                    'heating_energy_ref': 8000*(10*60), # From W to J in a timestep ot 10 minutes: P*t
                    "beta": 0.001,
                    "rho_1": 0.001,  # Weight for thermal comfort
                    "rho_2": 0.999,  # Weight for indoor air quality
                },
            ),
        ),
    }
)

eprllib_config.episodes(
    episode_fn = task_cofiguration,
    # read the json config file as a dict.
    episode_fn_config = episode_config,
)

assert eprllib_config.agents_config is not None, "Agents configuration is not defined."

number_of_agents = len([keys for keys in eprllib_config.agents_config.keys()])
ray.init(_temp_dir='C:/Users/grhen/ray_results/tmp')
register_env(name="EPEnv", env_creator=lambda args: Environment(args))
env_config = eprllib_config.to_dict()

if not restore:
    algo = PPOConfig()
    algo.training(
        
        # === General Algo Configs ===
        gamma = 0.99,
        lr_schedule = [
                (0, 3e-4),
                (2e6, 1e-4),
                (4e6, 5e-5)
                ],
        train_batch_size = 50000,
        minibatch_size = 2000,
        num_epochs = 30,
        
        # === Policy Model configuration ===
        model = {
            # FC Hidden layers
            "fcnet_hiddens": [128,128,128],
            "fcnet_activation": "tanh",
        },
        # === PPO Configs ===
        use_critic = True,
        use_gae = True,
        lambda_ = 0.92,
        use_kl_loss = True,
        kl_coeff = 0.2,
        kl_target = 0.7,
        shuffle_batch_per_epoch = True,
        vf_loss_coeff = 0.5,
        entropy_coeff_schedule = [
                (0, 0.02),
                (2e6, 0.005),
                (4e6, 0.001)
                ],
        clip_param = 0.25,
        vf_clip_param = 0.25,
    )
    algo.learners(
        num_learners = 0,
        num_cpus_per_learner = 1,
    )
    algo.environment(
        env = "EPEnv",
        env_config = env_config,
    )
    algo.framework(
        framework = 'torch',
    )
    algo.fault_tolerance(
        restart_failed_env_runners = True,
    )
    algo.env_runners(
        num_env_runners = 7,
        num_envs_per_env_runner = 1,
        sample_timeout_s = 100000,
        create_env_on_local_worker = True,
        rollout_fragment_length = 'auto',
        batch_mode = "complete_episodes",    
        explore = True,        
        observation_filter = "MeanStdFilter",
    )
    algo.multi_agent(
        policies = {
            'single_policy': PolicySpec(),
        },
        policy_mapping_fn = policy_map_fn,
        count_steps_by = "env_steps",
    )
    algo.reporting(
        min_sample_timesteps_per_iteration = 50,
    )
    algo.checkpointing(
        export_native_model_files = True,
    )
    algo.debugging(
        log_level = "ERROR",
        seed = 1,
    )
    algo.resources(
        num_gpus = 0,
    )
    algo.api_stack(
    enable_rl_module_and_learner=False,
    enable_env_runner_and_connector_v2=False,
)

    tuning = tune.Tuner(
        "PPO",
        param_space = algo.to_dict(),
        tune_config=tune.TuneConfig(
            num_samples = 1,
            reuse_actors = False,
            trial_name_creator = lambda trial: "{}_{}_{}".format(name, trial.trainable_name, trial.trial_id),
            trial_dirname_creator = lambda trial: "{}_{}_{}".format(name, trial.trainable_name, trial.trial_id),
        ),
        run_config=air.RunConfig(
            name = "{date}_{name}_{algorithm}".format(
                date = time.strftime("%Y%m%d%H%M%S"),
                name = name,
                algorithm = "PPO",
            ),
            storage_path = f'C:/Users/grhen/ray_results/{experiment_name}',
            stop = {"info/num_env_steps_trained": 1008*200000},
            log_to_file = True,
            
            checkpoint_config=air.CheckpointConfig(
                checkpoint_at_end = True,
                checkpoint_frequency = 5,
            ),
            failure_config=air.FailureConfig(
                max_failures = 100,
            ),
        ),
    )
    tuning.fit()

else:
    tuning = tune.Tuner.restore(checkpoint_path, 'PPO')
    tuning.fit()
    


"""
Defining agents
================

This module implements the classes to define agents. Agents are defined by the ``AgentSpec`` class. This class
contains the observation, filter, action, trigger, and reward specifications. The observation is defined by the
``ObservationSpec`` class. The filter is defined by the ``FilterSpec`` class. The action is defined by the ``ActionSpec`` class.
The trigger is defined by the ``TriggerSpec`` class. The reward is defined by the ``RewardSpec`` class.

The ``AgentSpec`` class has a method called ``build`` that is used to build the ``AgentSpec`` object. This method is used to
validate the properties of the object and to return the object as a dictionary. It is used internally when you build
the environment to provide it to RLlib.
"""
from typing import Dict, Any, Optional
from eprllib.Agents.Rewards.RewardSpec import RewardSpec
from eprllib.Agents.Filters.FilterSpec import FilterSpec
from eprllib.Agents.ActionSpec import ActionSpec
from eprllib.Agents.Triggers.TriggerSpec import TriggerSpec
from eprllib.Agents.ObservationSpec import ObservationSpec
from eprllib import logger

class AgentSpec:
    """
    AgentSpec is the base class for an agent specification to safe configuration of the object.
    """
    def __init__(
        self,
        observation: Optional[ObservationSpec] = None,
        filter: Optional[FilterSpec] = None,
        action: Optional[ActionSpec] = None,
        trigger: Optional[TriggerSpec] = None,
        reward: Optional[RewardSpec] = None,
    **kwargs: Any) -> None:
        """
        Contruction method for the AgentSpec class.

        Args:
            observation (ObservationSpec, optional): Defines the observation of the agent using
            the ObservationSpec class or a Dict. Defaults to NotImplemented.
            filter (FilterSpec, optional): Defines the filter for this agent using FilterSpec or a Dict. Defaults to None.
            action (ActionSpec, optional): Defines the action characteristics of the agent using ActionSpec or a Dict. Defaults to NotImplemented.
            trigger (TriggerSpec, optional): Defines the trigger for the agent using TriggerSpec or a Dict. Defaults to None.
            reward (RewardSpec, optional): Defines the reward elements of the agent using RewardSpec or a Dict. Defaults to NotImplemented.

        Raises:
            NotImplementedError: _description_
            NotImplementedError: _description_
            NotImplementedError: _description_
        """
        if observation is None:
            logger.info("No observation defined. Using default observation.")
            self.observation: ObservationSpec|Dict[str, Any] = ObservationSpec()
        else:
            self.observation = observation
        if filter is None:
            logger.info("No filter defined. Using default filter.")
            self.filter: FilterSpec|Dict[str, Any] = FilterSpec()
        else:
            self.filter = filter
        if action is None:
            logger.info("No action defined. Using default action.")
            self.action: ActionSpec|Dict[str, Any] = ActionSpec()
        else:
            self.action = action
        if trigger is None:
            logger.info("No trigger defined. Using default trigger.")
            self.trigger: TriggerSpec|Dict[str, Any] = TriggerSpec()
        else:
            self.trigger = trigger
        if reward is None:
            logger.info("No reward defined. Using default reward.")
            self.reward: RewardSpec|Dict[str, Any] = RewardSpec()
        else:
            self.reward = reward
        
        for key, value in kwargs.items():
            setattr(self, key, value)
    
    def __getitem__(self, key:str):
        return getattr(self, key)

    def __setitem__(self, key:str, value:Any):
        setattr(self, key, value)
        
        
    def build(self) -> Dict[str, Any]:
        """
        This method is used to build the AgentSpec object.
        """
        if isinstance(self.observation, ObservationSpec):
            self.observation = self.observation.build()
        else:
            logger.error(f"The observation must be defined as an ObservationSpec object but {type(self.observation)} was given.")
            raise ValueError(f"The observation must be defined as an ObservationSpec object but {type(self.observation)} was given.")
        if isinstance(self.filter, FilterSpec):
            self.filter = self.filter.build()
        else:
            logger.error(f"The filter must be defined as a FilterSpec object but {type(self.filter)} was given.")
            raise ValueError(f"The filter must be defined as a FilterSpec object but {type(self.filter)} was given.")
        if isinstance(self.action, ActionSpec):
            self.action = self.action.build()
        else:
            logger.error(f"The action must be defined as an ActionSpec object but {type(self.action)} was given.")
            raise ValueError(f"The action must be defined as an ActionSpec object but {type(self.action)} was given.")
        if isinstance(self.trigger, TriggerSpec):
            self.trigger = self.trigger.build()
        else:
            logger.error(f"The trigger must be defined as a TriggerSpec object but {type(self.trigger)} was given.")
            raise ValueError(f"The trigger must be defined as a TriggerSpec object but {type(self.trigger)} was given.")
        if isinstance(self.reward, RewardSpec):
            self.reward = self.reward.build()
        else:
            logger.error(f"The reward must be defined as a RewardSpec object but {type(self.reward)} was given.")
            raise ValueError(f"The reward must be defined as a RewardSpec object but {type(self.reward)} was given.")

        return vars(self)



"""
Utility functions for post-processing experience in RLlib environments
========================================================================

This module provides a function to generate experience from an environment and store it in a pandas DataFrame.
It supports both single-agent and multi-agent environments, with optional RNN state handling.

Functions
----------
generate_experience(env, env_config, agent_list, policies, rnn_use, num_episodes=1)
"""

from typing import List, Dict, Any
import pandas as pd
from ray.rllib.policy.policy import Policy
from eprllib.Environment.Environment import Environment
from eprllib import logger


def generate_experience(
    env: Environment,
    env_config: Dict[str,Any],
    agent_list: List[str],
    policies: Dict[str,Policy],
    rnn_use: Dict[str, bool],
    num_episodes: int = 1
) -> pd.DataFrame:
    """Function to generate experience and store in a DataFrame

    Args:
        env (Environment): The environment to use.
        env_config (Dict[str,Any]): The configuration for the environment.
        agent_list (List[str]): List of agents in the environment.
        policies (Dict[str,Policy]): Dictionary of policies for each agent.
        rnn_use (Dict[str, bool]): Dictionary indicating if RNN is used for each agent.
        num_episodes (int, optional): Number of episodes to generate. Defaults to 1.

    Returns:
        pd.DataFrame: DataFrame containing the experience of all episodes for each agent.
    """
    env = env(env_config)
    # Experience storage.
    all_observations = {agent: [] for agent in agent_list}
    all_actions = {agent: [] for agent in agent_list}
    all_rewards = {agent: [] for agent in agent_list}
    all_dones = {agent: [] for agent in agent_list}
    all_infos = {agent: [] for agent in agent_list}
    if rnn_use:
        all_states = {agent: [] for agent in agent_list}
    
    for _ in range(num_episodes):
        
        # Episode experience storage.
        episode_observations = {agent: [] for agent in agent_list}
        episode_actions = {agent: [] for agent in agent_list}
        episode_rewards = {agent: [] for agent in agent_list}
        episode_dones = {agent: [] for agent in agent_list}
        episode_infos = {agent: [] for agent in agent_list}
        if rnn_use:
            episode_states = {agent: [] for agent in agent_list}
            state = {agent: policies[agent].get_initial_state() for agent in agent_list}

        # Reset the environment and get the initial observation and infos.
        obs, info = env.reset()

        # List the agents in the environment for the first timestep.
        _agent_ids = list(obs.keys())

        # Reward for timestep 0.
        reward = {agent: 0 for agent in _agent_ids}
        # Control variable to finish the episode.
        done = {"__all__": False}
        
        # Store the experience for each agent.
        for agent in _agent_ids:
            episode_observations[agent].append(obs[agent])
            episode_rewards[agent].append(reward[agent])
            episode_dones[agent].append(done["__all__"])
            episode_infos[agent].append(info[agent])
            if rnn_use:
                episode_states[agent].append(state[agent])
        
        # Loop to generate the experience.
        while not done["__all__"]:
            # Action dictionary.
            action: Dict[str,Any] = {agent: 0 for agent in _agent_ids}
            if not rnn_use:
                # Get the action for each agent in the timestep.
                for agent in _agent_ids:
                    action[agent], _, _ = policies[agent].compute_single_action(obs[agent], explore=False) # No exploration
                    episode_actions[agent].append(action[agent])
            else:
                # Get the action for each agent in the timestep.
                for agent in _agent_ids:
                    action[agent], state[agent], _ = policies[agent].compute_single_action(obs[agent], state=state[agent], explore=False) # No exploration
                    episode_actions[agent].append(action[agent])
            # Execute the action in the environment.   
            new_obs, new_reward, done, truncateds, new_info = env.step(action)

            # Update the observation and info for the next timestep.
            obs = new_obs
            info = new_info
            reward = new_reward
            # List the agents in the environment for the next timestep.
            _agent_ids = list(obs.keys())
            
            # Store the experience for each agent.
            for agent in _agent_ids:
                episode_observations[agent].append(obs[agent])
                episode_rewards[agent].append(reward[agent])
                episode_dones[agent].append(done["__all__"])
                episode_infos[agent].append(info[agent])
                if rnn_use:
                    episode_states[agent].append(state[agent])
            

        # Store the episode experience.
        for agent in agent_list:
            all_observations[agent].extend(episode_observations[agent])
            all_actions[agent].extend(episode_actions[agent])
            all_rewards[agent].extend(episode_rewards[agent])
            all_dones[agent].extend(episode_dones[agent])
            all_infos[agent].extend(episode_infos[agent])
            if rnn_use:
                all_states[agent].extend(episode_states[agent])

    # Create the DataFrame with the experience of all episodes.
    if not rnn_use:
        return pd.DataFrame({
            agent:{
                "observations": all_observations[agent],
                "actions": all_actions[agent],
                "rewards": all_rewards[agent],
                "dones": all_dones[agent],
                "infos": all_infos[agent]
            }
            for agent
            in agent_list
        })
    else:
        return pd.DataFrame({
            agent:{
                "observations": all_observations[agent],
                "actions": all_actions[agent],
                "rewards": all_rewards[agent],
                "dones": all_dones[agent],
                "infos": all_infos[agent],
                "states": all_states[agent]
            }
            for agent
            in agent_list
        })
        


"""
Hierarchical trigger
=====================

This class uses a discrete action space. The size of the action space must be specified in the 
`trigger_fn_config` dictionary with the key "action_space_dim".
"""
from gymnasium import Space
from gymnasium.spaces import Discrete, MultiDiscrete
from typing import Any, List, Dict
from eprllib.Agents.Triggers.BaseTrigger import BaseTrigger
from eprllib.Utils.annotations import override
from eprllib.Utils.agent_utils import config_validation
from eprllib import logger

class HierarchicalGoalTriggerDiscrete(BaseTrigger):
    REQUIRED_KEYS: Dict[str, Any] = {
        "action_space_dim": int,
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any] = {}
    ):
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
    
    @override(BaseTrigger)
    def get_action_space_dim(self) -> Space[Any]:
        """This method is used to get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return Discrete(self.trigger_fn_config['action_space_dim'])

    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]):
        """
        This method is not used in top_level_agent.
        """
        msg = "This method should not be called."
        logger.error(msg)
        raise ValueError(msg)
    
    @override(BaseTrigger)
    def get_actuator_action(self, action: float | int, actuator: str):
        """
        This method is not used in top_level_agent.
        """
        msg = f"The actuator {actuator} is not used in the top_level_agent."
        logger.error(msg)
        raise ValueError(msg)

    @override(BaseTrigger)
    def action_to_goal(self, action: int | float) -> int | float:
        """
        This method is used to transform the action to a goal. The goal is used to define the reward.

        Args:
            action (Any): The action to be transformed.

        Returns:
            Any: The transformed action.
        """
        return action/(self.trigger_fn_config['action_space_dim']-1)
    
    
class HierarchicalObjectiveTriggerMultiDiscrete(BaseTrigger):
    REQUIRED_KEYS = {
        "action_space_dim": int,
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any] = {}
    ):
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
    
    @override(BaseTrigger)
    def get_action_space_dim(self) -> Space[Any]:
        """This method is used to get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        # TODO: Verify that the action space is the espected.
        # The action space is a MultiDiscrete space with the same dimension for each agent.
        return MultiDiscrete([self.trigger_fn_config['action_space_dim']] * 10)

    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]):
        """
        This method is not used in top_level_agent.
        """
        msg = "This method should not be called."
        logger.error(msg)
        raise ValueError(msg)
    
    @override(BaseTrigger)
    def get_actuator_action(self, action: float | int, actuator: str):
        """
        This method is not used in top_level_agent.
        """
        msg = f"The actuator {actuator} is not used in the top_level_agent."
        logger.error(msg)
        raise ValueError(msg)

    @override(BaseTrigger)
    def action_to_goal(self, action: int | float) -> int | float:
        """
        This method is used to transform the action to a goal. In this case, the agents are using a MultiDiscrete action 
        space, that are transformed to a single vector.

        Args:
            action (Any): The action to be transformed.

        Returns:
            Any: The transformed action.
        """
        return action
    


"""
Agents Connector Base Method
=============================

This module defines the base class for connector functions that allow the combination of 
agents' observations to provide a flexible configuration of the communication between agents. 
Built-in hierarchical (only two levels), fully-shared, centralized, and independent configurations 
are provided.
"""
from typing import Dict, Any, Tuple
from gymnasium import spaces
from eprllib import logger

class BaseConnector:
    def __init__(
        self,
        connector_fn_config: Dict[str, Any] = {}
    ):
        """
        Base class for connector functions.
        
        :param connector_fn_config: Configuration of the connector function.
        :type connector_fn_config: Dict[str, Any], optional
        """
        self.connector_fn_config = connector_fn_config
        self.obs_indexed: Dict[str,Dict[str, int]] = {}
    
    def __name__(self):
        """
        Returns the name of the connector function.

        :return: Name of the connector function.
        :rtype: str
        """
        return self.__class__.__name__
    
    def get_agent_obs_dim(
        self,
        env_config: Dict[str, Any],
        agent: str
    ) -> spaces.Space[Any]:
        """
        Get the agent observation dimension.

        :param env_config: Environment configuration.
        :type env_config: Dict[str, Any]
        :param agent: Agent identifier, optional.
        :type agent: str, optional
        :return: Agent observation spaces.
        :rtype: gym.spaces.Space
        """
        msg = "This method must be implemented in the child class."
        logger.error(msg)
        raise NotImplementedError(msg)
    
    def get_agent_obs_indexed(
        self,
        env_config: Dict[str, Any],
        agent: str
    ) -> Dict[str, int]:
        """
        Get a dictionary of the agent observation parameters and their respective index in the observation array.

        :param env_config: Environment configuration.
        :type env_config: Dict[str, Any]
        :param agent: Agent identifier, optional.
        :type agent: str, optional
        :return: Agent observation spaces.
        :rtype: gym.spaces.Space
        """
        msg = "This method must be implemented in the child class."
        logger.error(msg)
        raise NotImplementedError(msg)
    
    def get_all_agents_obs_spaces_dict(
        self,
        env_config: Dict[str, Any],
    ) -> spaces.Dict:
        """
        Get all the agents observations spaces putting togheter in a Dict space dimension.

        :param env_config: Environment configuration.
        :type env_config: Dict[str, Any]
        :return: Agents observation spaces.
        :rtype: gym.spaces.Dict
        """
        possible_agents = [key for key in env_config["agents_config"].keys()]
        observation_space_dict: Dict[str, Any] = {}
        for agent in possible_agents:
            observation_space_dict[agent] = self.get_agent_obs_dim(env_config, agent)
        return spaces.Dict(observation_space_dict)
        
    def set_top_level_obs(
        self,
        env_config: Dict[str, Any],
        agent_states: Dict[str, Dict[str, Any]],
        dict_agents_obs: Dict[str, Any],
        infos: Dict[str, Dict[str, Any]],
        is_last_timestep: bool = False
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]:
        """
        Set the multi-agent observation.

        :param env_config: Environment configuration.
        :type env_config: Dict[str, Any]
        :param agent_states: Agent states.
        :type agent_states: Dict[str, Dict[str, Any]]
        :param dict_agents_obs: Dictionary of agents' observations.
        :type dict_agents_obs: Dict[str, Any]
        :param infos: Additional information.
        :type infos: Dict[str, Dict[str, Any]]
        :param is_last_timestep: Flag indicating if it is the last timestep, defaults to False.
        :type is_last_timestep: bool, optional
        :return: Multi-agent observation, updated infos, and a flag indicating if it is the lowest level.
        :rtype: Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]
        """
        is_lowest_level = True
        return dict_agents_obs, infos, is_lowest_level
    
    def set_low_level_obs(
        self,
        env_config: Dict[str, Any],
        agent_states: Dict[str,Dict[str,Any]],
        dict_agents_obs: Dict[str,Any],
        infos: Dict[str, Dict[str, Any]],
        goals: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]:
        """
        Set the multiagent observation.

        :param env_config: environment configuration
        :type env_config: Dict[str,Any]
        :param agent_states: agent states
        :type agent_states: Dict[str,Any]
        :param dict_agents_obs: dictionary of agents observations
        :type dict_agents_obs: Dict[str,Any]
        :return: multiagent observation
        :rtype: Dict[str,Any]
        """
        is_lowest_level = True
        return dict_agents_obs, infos, is_lowest_level



"""
Default Agents Connector
=========================

This module defines the default connector class that allows the combination of agents' observations 
to provide a flexible configuration of the communication between agents. Built-in hierarchical 
(only two levels), fully-shared, centralized, and independent configurations are provided.
"""
from typing import Dict, Any, Tuple
from gymnasium.spaces import Box
from eprllib.AgentsConnectors.BaseConnector import BaseConnector
from eprllib.Utils.annotations import override
from eprllib.Utils.connector_utils import (
    set_variables_in_obs,
    set_internal_variables_in_obs,
    set_meters_in_obs,
    set_simulation_parameters_in_obs,
    set_zone_simulation_parameters_in_obs,
    set_prediction_variables_in_obs,
    set_other_obs_in_obs,
    set_actuators_in_obs,
    set_user_occupation_forecast_in_obs
    )
from eprllib import logger

class DefaultConnector(BaseConnector):
    def __init__(
        self,
        connector_fn_config: Dict[str,Any] = {}
    ):
        """
        Base class for multiagent functions.
        
        :param connector_fn_config: configuration of the multiagent function
        :type connector_fn_config: Dict[str,Any], optional
        """
        super().__init__(connector_fn_config)
    
    @override(BaseConnector)
    def get_agent_obs_dim(
        self,
        env_config: Dict[str,Any],
        agent: str
        ) -> Box:
        """
        Get the agent observation dimension.

        :param env_config: environment configuration
        :type env_config: Dict[str,Any]
        :return: agent observation spaces
        :rtype: Dict[str, gym.Space]
        """
        obs_space_len: int = 0
        self.obs_indexed[agent] = {}
        
        self.obs_indexed[agent], obs_space_len = set_variables_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_internal_variables_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_meters_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_simulation_parameters_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_zone_simulation_parameters_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_prediction_variables_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_other_obs_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_actuators_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_user_occupation_forecast_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
                
        assert obs_space_len > 0, "The observation space length must be greater than 0."
        assert len(self.obs_indexed[agent]) == obs_space_len, f"The observation space length must be equal to the number of indexed observations. Obs indexed:{len(self.obs_indexed[agent])} != Obs space len:{obs_space_len}."
        # obs_space_len += 1
        logger.debug(f"Observation space length for agent {agent}: {obs_space_len}")
        
        return Box(float("-inf"), float("inf"), (obs_space_len, ))
    
    @override(BaseConnector)
    def get_agent_obs_indexed(
        self,
        env_config: Dict[str, Any],
        agent: str
    ) -> Dict[str, int]:
        """
        Get a dictionary of the agent observation parameters and their respective index in the observation array.

        :param env_config: Environment configuration.
        :type env_config: Dict[str, Any]
        :param agent: Agent identifier, optional.
        :type agent: str, optional
        :return: Agent observation spaces.
        :rtype: gym.spaces.Space
        """
        if self.obs_indexed == {}:
            self.get_agent_obs_dim(env_config, agent)
        return self.obs_indexed[agent]
    
    @override(BaseConnector)    
    def set_top_level_obs(
        self,
        env_config: Dict[str, Any],
        agent_states: Dict[str,Dict[str,Any]],
        dict_agents_obs: Dict[str,Any],
        infos: Dict[str, Dict[str, Any]],
        is_last_timestep: bool = False
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]:
        """
        Set the multiagent observation.

        :param env_config: environment configuration
        :type env_config: Dict[str,Any]
        :param agent_states: agent states
        :type agent_states: Dict[str,Any]
        :param dict_agents_obs: dictionary of agents observations
        :type dict_agents_obs: Dict[str,Any]
        :return: multiagent observation
        :rtype: Dict[str,Any]
        """
        is_lowest_level = True
        return dict_agents_obs, infos, is_lowest_level
    
    @override(BaseConnector)
    def set_low_level_obs(
        self,
        env_config: Dict[str, Any],
        agent_states: Dict[str,Dict[str,Any]],
        dict_agents_obs: Dict[str,Any],
        infos: Dict[str, Dict[str, Any]],
        goals: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]:
        """
        Set the multiagent observation.

        :param env_config: environment configuration
        :type env_config: Dict[str,Any]
        :param agent_states: agent states
        :type agent_states: Dict[str,Any]
        :param dict_agents_obs: dictionary of agents observations
        :type dict_agents_obs: Dict[str,Any]
        :return: multiagent observation
        :rtype: Dict[str,Any]
        """
        is_lowest_level = True
        return dict_agents_obs, infos, is_lowest_level



"""
Independent Agents Connector
=============================

This module implements the default observation function where each agent has its own observation space 
and it is returned without modifications, considering only the agent_states provided in the BaseRunner class.
"""

from typing import Any, Dict
from eprllib.AgentsConnectors.DefaultConnector import DefaultConnector
from eprllib import logger

class IndependentConnector(DefaultConnector):
    def __init__(
        self,
        connector_fn_config: Dict[str, Any] = {}
    ):
        """
        Initializes the independent connector.

        Args:
            connector_fn_config (Dict[str, Any]): Configuration dictionary for the multi-agent function.
        """
        super().__init__(connector_fn_config)
        logger.debug("IndependentConnector initialized with configuration: %s", connector_fn_config)


from eprllib.Agents.Filters.FullySharedParametersFilter import FullySharedParametersFilter
from typing import Any, Dict
import numpy as np
import pytest

class TestFullysharedparametersfilter:

    def test___init___1(self):
        """
        Test the initialization of FullySharedParametersFilter.

        This test verifies that the FullySharedParametersFilter is correctly initialized
        with the provided filter_fn_config and that the agent_name attribute is set to None.
        """
        filter_fn_config: Dict[str, Any] = {}
        filter_instance = FullySharedParametersFilter(filter_fn_config)

        assert isinstance(filter_instance, FullySharedParametersFilter)
        assert filter_instance.agent_name is None

    def test_get_filtered_obs_1(self):
        """
        Test that get_filtered_obs correctly filters out actuator states when self.agent_name is None.

        This test verifies that:
        1. The method correctly identifies the agent name from the agent_states dictionary.
        2. Actuator states are removed from the observation.
        3. The returned observation is a numpy array of float32 values.
        4. The returned observation contains only the non-actuator state values.
        """
        # Setup
        filter_fn_config = {}
        env_config = {
            "agents_config": {
                "agent1": {
                    "action": {
                        "actuators": [
                            ["actuator1", "type1", "subtype1"],
                            ["actuator2", "type2", "subtype2"]
                        ]
                    }
                }
            }
        }
        agent_states = {
            "agent1: state1": 1.0,
            "agent1: state2": 2.0,
            "agent1: actuator1: type1: subtype1": 3.0,
            "agent1: actuator2: type2: subtype2": 4.0
        }

        # Create filter instance
        filter_instance = FullySharedParametersFilter(filter_fn_config)

        # Call the method under test
        result = filter_instance.get_filtered_obs(env_config, agent_states)

        # Assertions
        assert isinstance(result, np.ndarray)
        assert result.dtype == np.float32
        np.testing.assert_array_equal(result, np.array([1.0, 2.0], dtype=np.float32))
        assert filter_instance.agent_name == "agent1"

    def test_get_filtered_obs_2(self):
        """
        Test the get_filtered_obs method when self.agent_name is not None.
        This test verifies that the method correctly filters out actuator states
        and returns a numpy array of the remaining agent states.
        """
        filter_fn_config = {}
        filter_instance = FullySharedParametersFilter(filter_fn_config)
        filter_instance.agent_name = "agent1"

        env_config = {
            "agents_config": {
                "agent1": {
                    "action": {
                        "actuators": [
                            ["actuator1", "type1", "subtype1"],
                            ["actuator2", "type2", "subtype2"]
                        ]
                    }
                }
            }
        }

        agent_states = {
            "agent1: state1": 1.0,
            "agent1: state2": 2.0,
            "agent1: actuator1: type1: subtype1": 3.0,
            "agent1: actuator2: type2: subtype2": 4.0,
            "agent1: state3": 5.0
        }

        result = filter_instance.get_filtered_obs(env_config, agent_states)

        expected_result = np.array([1.0, 2.0, 5.0], dtype='float32')
        np.testing.assert_array_equal(result, expected_result)

    def test_get_filtered_obs_empty_agent_states(self):
        """
        Test the get_filtered_obs method with empty agent_states.
        This tests the edge case where no agent states are provided, which should result in an empty numpy array.
        """
        filter_fn_config = {}
        filter_instance = FullySharedParametersFilter(filter_fn_config)

        env_config = {"agents_config": {"agent1": {"action": {"actuators": []}}}}
        agent_states = {}

        with pytest.raises(ValueError):
            filter_instance.get_filtered_obs(env_config, agent_states)
        

    def test_get_filtered_obs_missing_agent_config(self):
        """
        Test the get_filtered_obs method when the agent configuration is missing from the env_config.
        This tests the edge case where the expected agent configuration is not present in the environment config.
        """
        filter_fn_config = {}
        filter_instance = FullySharedParametersFilter(filter_fn_config)

        env_config = {"agents_config": {}}
        agent_states = {"agent1: state1": 1.0, "agent1: state2": 2.0}

        with pytest.raises(KeyError):
            filter_instance.get_filtered_obs(env_config, agent_states)

    def test_get_filtered_obs_no_actuators(self):
        """
        Test the get_filtered_obs method when there are no actuators configured for the agent.
        This tests the edge case where the agent has no actuators, so no filtering should occur.
        """
        filter_fn_config = {}
        filter_instance = FullySharedParametersFilter(filter_fn_config)

        env_config = {"agents_config": {"agent1": {"action": {"actuators": []}}}}
        agent_states = {"agent1: state1": 1.0, "agent1: state2": 2.0}

        result = filter_instance.get_filtered_obs(env_config, agent_states)
        expected = np.array([1.0, 2.0], dtype=np.float32)
        np.testing.assert_array_equal(result, expected)

    def test_init_with_empty_dict(self):
        """
        Test initialization of FullySharedParametersFilter with an empty dictionary.
        This is to verify that the method handles the edge case of an empty input correctly.
        """
        filter_fn_config = {}
        filter_instance = FullySharedParametersFilter(filter_fn_config)
        assert filter_instance.agent_name is None
        assert isinstance(filter_instance, FullySharedParametersFilter)



"""
Setpoint triggers
==================

This module contains classes to implement setpoint triggers for controlling actuators in the environment.
"""
import gymnasium as gym
import numpy as np
from typing import Any, Dict, List, Tuple, Optional
from eprllib.Agents.Triggers.BaseTrigger import BaseTrigger
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override
from eprllib.Utils.agent_utils import get_agent_name, config_validation
from eprllib import logger

class DualSetpointTriggerDiscreteAndAvailabilityTrigger(BaseTrigger):
    REQUIRED_KEYS: Dict[str, Any] = {
        "temperature_range": Tuple[int|float, int|float],
        "actuator_for_cooling": Tuple[str, str, str],
        "actuator_for_heating": Tuple[str, str, str],
        "availability_actuator": Tuple[str, str, str],
        "action_space_dim": int
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any]
    ):
        """
        This class implements the Dual Set Point Thermostat action function.

        Args:
            trigger_fn_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - temperature_range (Tuple[int, int]): The temperature range for the setpoints.
                - actuator_for_cooling (Tuple[str, str, str]): The configuration for the cooling actuator.
                - actuator_for_heating (Tuple[str, str, str]): The configuration for the heating actuator.
                - availability_actuator (Tuple[str, str, str]): The configuration for the availability actuator.
                - action_space_dim (int): The dimension of the action space. Must be greater than or equal to 3.
        """
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
        
        self.agent_name: Optional[str] = None
        temperature_range: Tuple[float|int, float|int] = trigger_fn_config['temperature_range']
        self.temperature_range_max: float|int = max(temperature_range)
        self.temperature_range_min: float|int = min(temperature_range)
        self.temperature_range_avg: float = (self.temperature_range_max + self.temperature_range_min) / 2
        self.actuator_for_cooling = None
        self.actuator_for_heating = None
        self.availability_actuator = None
        # self.band_gap_range_len: int = trigger_fn_config['band_gap_range_len'] + 1
        self.action_space_dim: int = trigger_fn_config['action_space_dim']
        if self.action_space_dim < 3:
            raise ValueError("The action_space_dim must be greater than or equal to 3.")
    
    @override(BaseTrigger)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Discrete(self.action_space_dim)
    
    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """
        if self.agent_name is None:
            self.agent_name = get_agent_name(actuators)
            self.actuator_for_cooling = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['actuator_for_cooling'][0],
                self.trigger_fn_config['actuator_for_cooling'][1],
                self.trigger_fn_config['actuator_for_cooling'][2]
            )
            self.actuator_for_heating = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['actuator_for_heating'][0],
                self.trigger_fn_config['actuator_for_heating'][1],
                self.trigger_fn_config['actuator_for_heating'][2]
            )
            self.availability_actuator = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['availability_actuator'][0],
                self.trigger_fn_config['availability_actuator'][1],
                self.trigger_fn_config['availability_actuator'][2]
            )
            
        actuator_dict_actions: Dict[str, Any] = {actuator: None for actuator in actuators}
        
        assert self.actuator_for_cooling is not None, "Cooling actuator name has not been initialized."
        assert self.actuator_for_heating is not None, "Heating actuator name has not been initialized."
        assert self.availability_actuator is not None, "Availability actuator name has not been initialized."
        
        if action == 0:
            actuator_dict_actions.update({
                self.actuator_for_cooling: self.temperature_range_avg + 1,
                self.actuator_for_heating: self.temperature_range_avg,
                self.availability_actuator: 0
            })
        
        else:
            actuator_dict_actions.update({
                self.actuator_for_cooling: self.temperature_range_avg + (((action-1)/(self.action_space_dim-2))*(self.temperature_range_max - 1 - self.temperature_range_avg)) + 1,
                self.actuator_for_heating: self.temperature_range_avg - (((action-1)/(self.action_space_dim-2))*(self.temperature_range_avg - self.temperature_range_min)),
                self.availability_actuator: 1
            })

        # Check if there is an actuator_dict_actions value equal to None.
        for actuator in actuator_dict_actions:
            if actuator_dict_actions[actuator] is None:
                msg = f"The actuator {actuator} is not in the list of actuators: \n{actuators}.\nThe actuator dict is: \n{actuator_dict_actions}"
                logger.error(msg)
                raise ValueError(msg)
        
        return actuator_dict_actions
    
    @override(BaseTrigger)
    def get_actuator_action(self, action:float|int, actuator: str) -> Any:
        """
        This method is used to get the actions of the actuators after transform the 
        agent action to actuator action.

        Args:
            action (float|int): The action to be transformed.
            actuator (str): The actuator name.

        Returns:
            Any: The transformed action.
        """
        return action

class DualSetpointTriggerDiscreteAndAvailabilityTrigger_v2(BaseTrigger):
    REQUIRED_KEYS: Dict[str, Any] = {
        "temperature_range": Tuple[int|float, int|float],
        "actuator_for_cooling": Tuple[str, str, str],
        "actuator_for_heating": Tuple[str, str, str],
        "availability_actuator": Tuple[str, str, str],
        "action_space_dim": int
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any]
    ):
        """
        This class implements the Dual Set Point Thermostat action function.

        Args:
            trigger_fn_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - temperature_range (Tuple[int, int]): The temperature range for the setpoints.
                - actuator_for_cooling (Tuple[str, str, str]): The configuration for the cooling actuator.
                - actuator_for_heating (Tuple[str, str, str]): The configuration for the heating actuator.
                - availability_actuator (Tuple[str, str, str]): The configuration for the availability actuator.
                - action_space_dim (int): The dimension of the action space. Must be greater than or equal to 3.
        """
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
        
        self.agent_name: Optional[str] = None
        temperature_range: Tuple[float|int, float|int] = trigger_fn_config['temperature_range']
        self.temperature_range_max: float|int = max(temperature_range)
        self.temperature_range_min: float|int = min(temperature_range)
        self.actuator_for_cooling = None
        self.actuator_for_heating = None
        self.availability_actuator = None
        self.deadband: int = trigger_fn_config.get('deadband', 2)
        self.action_space_dim: int = trigger_fn_config['action_space_dim']
        if self.action_space_dim < 3:
            raise ValueError("The action_space_dim must be greater than or equal to 3.")
    
    @override(BaseTrigger)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Discrete(self.action_space_dim)
    
    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """
        if self.agent_name is None:
            self.agent_name = get_agent_name(actuators)
            self.actuator_for_cooling = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['actuator_for_cooling'][0],
                self.trigger_fn_config['actuator_for_cooling'][1],
                self.trigger_fn_config['actuator_for_cooling'][2]
            )
            self.actuator_for_heating = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['actuator_for_heating'][0],
                self.trigger_fn_config['actuator_for_heating'][1],
                self.trigger_fn_config['actuator_for_heating'][2]
            )
            self.availability_actuator = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['availability_actuator'][0],
                self.trigger_fn_config['availability_actuator'][1],
                self.trigger_fn_config['availability_actuator'][2]
            )
            
        actuator_dict_actions: Dict[str, Any] = {actuator: None for actuator in actuators}
        
        assert self.actuator_for_cooling is not None, "Cooling actuator name has not been initialized."
        assert self.actuator_for_heating is not None, "Heating actuator name has not been initialized."
        assert self.availability_actuator is not None, "Availability actuator name has not been initialized."
        
        if action == 0:
            actuator_dict_actions.update({
                self.actuator_for_heating: float(self.temperature_range_min),
                self.actuator_for_cooling: float(self.temperature_range_min + self.deadband),
                self.availability_actuator: 0
            })
        
        else:
            actuator_dict_actions.update({
                self.actuator_for_heating: float(self.temperature_range_min + ((action-1)/(self.action_space_dim-2))*(self.temperature_range_max - self.temperature_range_min - self.deadband)),
                self.actuator_for_cooling: float(self.temperature_range_min + ((action-1)/(self.action_space_dim-2))*(self.temperature_range_max - self.temperature_range_min - self.deadband) + self.deadband),
                self.availability_actuator: 1
            })

        # Check if there is an actuator_dict_actions value equal to None.
        for actuator in actuator_dict_actions:
            if actuator_dict_actions[actuator] is None:
                msg = f"The actuator {actuator} is not in the list of actuators: \n{actuators}.\nThe actuator dict is: \n{actuator_dict_actions}"
                logger.error(msg)
                raise ValueError(msg)
        
        return actuator_dict_actions
    
    @override(BaseTrigger)
    def get_actuator_action(self, action:float|int, actuator: str) -> Any:
        """
        This method is used to get the actions of the actuators after transform the 
        agent action to actuator action.

        Args:
            action (float|int): The action to be transformed.
            actuator (str): The actuator name.

        Returns:
            Any: The transformed action.
        """
        return action


class DualSetpointTriggerContinuosAndAvailabilityTrigger_v2(BaseTrigger):
    REQUIRED_KEYS: Dict[str, Any] = {
        "actuator_for_cooling": Tuple[str, str, str],
        "actuator_for_heating": Tuple[str, str, str],
        "availability_actuator": Tuple[str, str, str],
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any]
    ):
        """
        This class implements the Dual Set Point Thermostat action function.

        Args:
            trigger_fn_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - temperature_range (Tuple[int, int]): The temperature range for the setpoints.
                - actuator_for_cooling (Tuple[str, str, str]): The configuration for the cooling actuator.
                - actuator_for_heating (Tuple[str, str, str]): The configuration for the heating actuator.
                - availability_actuator (Tuple[str, str, str]): The configuration for the availability actuator.
                - action_space_dim (int): The dimension of the action space. Must be greater than or equal to 3.
        """
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
        
        
        temperature_range: Tuple[float|int, float|int] = trigger_fn_config.get('temperature_range', (16, 32))
        self.temperature_range_max: float = float(max(temperature_range))
        self.temperature_range_min: float = float(min(temperature_range))
        
        self.deadband: int = trigger_fn_config.get('deadband', 2)
        if (self.temperature_range_max - self.temperature_range_min) < self.deadband:
            raise ValueError(f"The temperature range must be at least {self.deadband} degrees wide to accommodate the deadband.")
        
        self.agent_name: Optional[str] = None
        self.actuator_for_cooling = None
        self.actuator_for_heating = None
        self.availability_actuator = None
    
    @override(BaseTrigger)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Dict({
            "availability": gym.spaces.Discrete(2), # 0: OFF, 1: ON
            "temperature": gym.spaces.Box(low=self.temperature_range_min, high=self.temperature_range_max, shape=(1,), dtype=np.float32)
        })
    
    
    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """
        if self.agent_name is None:
            self.agent_name = get_agent_name(actuators)
            self.actuator_for_cooling = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['actuator_for_cooling'][0],
                self.trigger_fn_config['actuator_for_cooling'][1],
                self.trigger_fn_config['actuator_for_cooling'][2]
            )
            self.actuator_for_heating = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['actuator_for_heating'][0],
                self.trigger_fn_config['actuator_for_heating'][1],
                self.trigger_fn_config['actuator_for_heating'][2]
            )
            self.availability_actuator = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['availability_actuator'][0],
                self.trigger_fn_config['availability_actuator'][1],
                self.trigger_fn_config['availability_actuator'][2]
            )
            
        actuator_dict_actions: Dict[str, Any] = {actuator: None for actuator in actuators}
        
        assert self.actuator_for_cooling is not None, "Cooling actuator name has not been initialized."
        assert self.actuator_for_heating is not None, "Heating actuator name has not been initialized."
        assert self.availability_actuator is not None, "Availability actuator name has not been initialized."
        
        # On-Off action
        if action["availability"] == 0:
            actuator_dict_actions.update({
                self.availability_actuator: 0
            })
        else:
            actuator_dict_actions.update({
                self.availability_actuator: 1
            })
        
        # Temperature actions
        actuator_dict_actions.update({
            self.actuator_for_cooling: action["temperature"][0] + self.deadband,
            self.actuator_for_heating: action["temperature"][0],
        })

        # Check if there is an actuator_dict_actions value equal to None.
        for actuator in actuator_dict_actions:
            if actuator_dict_actions[actuator] is None:
                msg = f"The actuator {actuator} is not in the list of actuators: \n{actuators}.\nThe actuator dict is: \n{actuator_dict_actions}"
                logger.error(msg)
                raise ValueError(msg)
        
        return actuator_dict_actions
    
    @override(BaseTrigger)
    def get_actuator_action(self, action:int|float, actuator: str) -> float|int:
        """
        This method is used to get the actions of the actuators after transform the 
        agent action to actuator action.

        Args:
            action (float|int): The action to be transformed.
            actuator (str): The actuator name.

        Returns:
            Any: The transformed action.
        """
        return action
    
class DualSetpointContinuosAndConstantFlowTrigger(BaseTrigger):
    REQUIRED_KEYS: Dict[str, Any] = {
        "actuator_for_cooling": Tuple[str, str, str],
        "actuator_for_heating": Tuple[str, str, str],
        # "availability_actuator": Tuple[str, str, str],
        # "air_mass_flow_rate": int|float,
        # "ideal_loads_air_system_actuator": Tuple[str, str, str],
        # "air_temperature_actuator": Tuple[str, str, str],
        # "air_temperature": int|float
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any]
    ):
        """
        This class implements the Dual Set Point Thermostat action function.

        Args:
            trigger_fn_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - temperature_range (Tuple[int, int]): The temperature range for the setpoints.
                - actuator_for_cooling (Tuple[str, str, str]): The configuration for the cooling actuator.
                - actuator_for_heating (Tuple[str, str, str]): The configuration for the heating actuator.
                - availability_actuator (Tuple[str, str, str]): The configuration for the availability actuator.
                - action_space_dim (int): The dimension of the action space. Must be greater than or equal to 3.
        """
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
        
        self.agent_name: Optional[str] = None
        
        temperature_range: Tuple[float|int, float|int] = trigger_fn_config.get('temperature_range', (16, 32))
        
        self.temperature_range_max: float|int = max(temperature_range)
        self.temperature_range_min: float|int = min(temperature_range)
        
        self.deadband: int = trigger_fn_config.get('deadband', 2)
        if (self.temperature_range_max - self.temperature_range_min) < self.deadband:
            raise ValueError(f"The temperature range must be at least {self.deadband} degrees wide to accommodate the deadband.")
        
        self.actuator_for_cooling = None
        self.actuator_for_heating = None
        
        # Create a .csv file to save the actions values during execution
        # self.actions_file_path: Optional[str] = trigger_fn_config.get('actions_file_path', None)
        # if self.actions_file_path is not None:
        #     self.actions_file = open(self.actions_file_path, 'w')

        
    @override(BaseTrigger)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32)
    
    
    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """
        if self.agent_name is None:
            self.agent_name = get_agent_name(actuators)
            self.actuator_for_cooling = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['actuator_for_cooling'][0],
                self.trigger_fn_config['actuator_for_cooling'][1],
                self.trigger_fn_config['actuator_for_cooling'][2]
            )
            self.actuator_for_heating = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['actuator_for_heating'][0],
                self.trigger_fn_config['actuator_for_heating'][1],
                self.trigger_fn_config['actuator_for_heating'][2]
            )
            
        actuator_dict_actions: Dict[str, Any] = {actuator: None for actuator in actuators}
        
        assert self.actuator_for_cooling is not None, "Cooling actuator name has not been initialized."
        assert self.actuator_for_heating is not None, "Heating actuator name has not been initialized."
        action = float(action) # Try to convert the action in to a float number.
        assert type(action) == float, f"The action must be a float but is {type(action)}."
        
        # Save the action to the file.
        # self.actions_file.write(f"{action}\n")
        # self.actions_file.flush()
        
        if action < 0.0:
            actuator_dict_actions.update({
                self.actuator_for_heating: float(self.temperature_range_min),
                self.actuator_for_cooling: float(self.temperature_range_min + self.deadband)
            })
        
        elif action > 1.0:
            actuator_dict_actions.update({
                self.actuator_for_heating: float(self.temperature_range_max - self.deadband),
                self.actuator_for_cooling: float(self.temperature_range_max)
            })
        
        else:
            actuator_dict_actions.update({
                self.actuator_for_heating: float(self.temperature_range_min + (action)*(self.temperature_range_max - self.temperature_range_min - self.deadband)),
                self.actuator_for_cooling: float(self.temperature_range_min + (action)*(self.temperature_range_max - self.temperature_range_min - self.deadband) + self.deadband)
            })

        # Check if there is an actuator_dict_actions value equal to None.
        for actuator in actuator_dict_actions:
            if actuator_dict_actions[actuator] is None:
                msg = f"The actuator {actuator} is not in the list of actuators: \n{actuators}.\nThe actuator dict is: \n{actuator_dict_actions}"
                logger.error(msg)
                raise ValueError(msg)
        
        return actuator_dict_actions
    
    @override(BaseTrigger)
    def get_actuator_action(self, action:int|float, actuator: str) -> float|int:
        """
        This method is used to get the actions of the actuators after transform the 
        agent action to actuator action.

        Args:
            action (float|int): The action to be transformed.
            actuator (str): The actuator name.

        Returns:
            Any: The transformed action.
        """
        return action
    

class AvailabilityTrigger(BaseTrigger):
    REQUIRED_KEYS = {
        "availability_actuator": Tuple[str, str, str]
    }
    
    def __init__(
        self, 
        trigger_fn_config:Dict[str,Any]
        ):
        """
        This class implements the Dual Set Point Thermostat action function.

        Args:
            trigger_fn_config (Dict[str,Any]): The configuration of the action function.
            It should contain the following keys: agents_type (Dict[str, int]): A dictionary 
            mapping agent names to their types (1 for cooling, 2 for heating, 3 for Availability).
        """
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
        
        self.agent_name: Optional[str] = None
        self.availability_actuator: Optional[str] = None
    
    @override(BaseTrigger)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Discrete(2)
    
    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """
        if self.agent_name is None:
            self.agent_name = get_agent_name(actuators)
            self.availability_actuator = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['availability_actuator'][0],
                self.trigger_fn_config['availability_actuator'][1],
                self.trigger_fn_config['availability_actuator'][2]
            )

        
        actuator_dict_actions: Dict[str, Any] = {actuator: None for actuator in actuators}
        
        assert self.availability_actuator is not None, "Availability actuator name has not been initialized."
        
        if action == 0:
            actuator_dict_actions.update({
                self.availability_actuator: 0
            })
        
        else:
            actuator_dict_actions.update({
                self.availability_actuator: 1
            })

        # Check if there is an actuator_dict_actions value equal to None.
        for actuator in actuator_dict_actions:
            if actuator_dict_actions[actuator] is None:
                msg = f"The actuator {actuator} is not in the list of actuators: \n{actuators}.\nThe actuator dict is: \n{actuator_dict_actions}"
                logger.error(msg)
                raise ValueError(msg)
        
        return actuator_dict_actions
    
    @override(BaseTrigger)
    def get_actuator_action(self, action:float|int, actuator: str) -> Any:
        """
        This method is used to get the actions of the actuators after transform the 
        agent action to actuator action.
        
        Args:
            action (float|int): The action to be transformed.
            actuator (str): The actuator name.
        
        Returns:
            Any: The transformed action.
        """
        return action

    @override(BaseTrigger)
    def action_to_goal(self, action: int | float) -> int | float:
        """
        This method is used to transform the action to a goal. The goal is used to define the reward.

        Args:
            action (Any): The action to be transformed.

        Returns:
            Any: The transformed action.
        """
        return action
    


"""
Windows Shading Triggers
========================

This module contains classes to implement window shading triggers for controlling actuators in the environment.
"""
import gymnasium as gym
from typing import Any, Dict, List, Tuple, Optional
from eprllib.Agents.Triggers.BaseTrigger import BaseTrigger
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override
from eprllib.Utils.agent_utils import get_agent_name, config_validation
from eprllib import logger

class WindowsShadingTrigger(BaseTrigger):
    REQUIRED_KEYS: Dict[str, Any] = {
        "shading_actuator": Tuple[str, str, str],
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any]
    ):
        """
        This class implements the window shading action function.

        Args:
            trigger_fn_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - shading_actuator (Tuple[str, str, str]): The configuration for the shading actuator.
        """
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
        
        self.agent_name: Optional[str] = None
        self.shading_actuator: Optional[str] = None
    
    @override(BaseTrigger)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Discrete(11)
    
    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str, Any]:
        """
        Transform the agent action to actuator action. Consider that one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int or float.
            actuators (List[str]): The list of actuator names that the agent manages.

        Returns:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """
        if self.agent_name is None:
            self.agent_name = get_agent_name(actuators)
            self.shading_actuator = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['shading_actuator'][0],
                self.trigger_fn_config['shading_actuator'][1],
                self.trigger_fn_config['shading_actuator'][2]
            )
        
        assert self.shading_actuator is not None, "Shading actuator name has not been initialized."
        
        actuator_dict_actions: Dict[str, Any] = {actuator: None for actuator in actuators}
        if action == 0:  # Shading device is off (applies to shades and blinds).
            actuator_dict_actions.update({self.shading_actuator: 0})
        else:  # 3.0: Exterior shade is on.
            actuator_dict_actions.update({self.shading_actuator: 3})
            
        # Check if there is an actuator_dict_actions value equal to None.
        for actuator in actuator_dict_actions:
            if actuator_dict_actions[actuator] is None:
                msg = f"The actuator {actuator} is not in the list of actuators: \n{actuators}.\nThe actuator dict is: \n{actuator_dict_actions}"
                logger.error(msg)
                raise ValueError(msg)
            
        return actuator_dict_actions
    
    @override(BaseTrigger)
    def get_actuator_action(self, action: float | int, actuator: str) -> Any:
        """
        Get the actions of the actuators after transforming the agent action to actuator action.

        Args:
            action (float | int): Action provided by the policy and transformed by agent_to_actuator_action.
            actuator (str): The actuator that requires the action.

        Returns:
            Any: The action for the actuator.
        """
        return action

    @override(BaseTrigger)
    def action_to_goal(self, action: int | float) -> int | float:
        """
        This method is used to transform the action to a goal. The goal is used to define the reward.

        Args:
            action (Any): The action to be transformed.

        Returns:
            Any: The transformed action.
        """
        return action
    


"""
Base Environment
=================

This script define the environment of EnergyPlus implemented in RLlib. To works 
need to define the EnergyPlus Runner.
"""
import tempfile
import collections
import numpy as np
from gymnasium import spaces
from ray.rllib.env.multi_agent_env import MultiAgentEnv
from queue import Empty, Full, Queue
from typing import Any, Dict, Optional, Tuple
from eprllib.Environment.EnvironmentConfig import EnvironmentConfig
from eprllib.Environment.EnvironmentRunner import EnvironmentRunner
from eprllib.Agents.Rewards.BaseReward import BaseReward
from eprllib.Agents.Filters.BaseFilter import BaseFilter
from eprllib.Agents.Triggers.BaseTrigger import BaseTrigger
from eprllib.AgentsConnectors.BaseConnector import BaseConnector
from eprllib.Episodes.BaseEpisode import BaseEpisode
from eprllib.Utils.annotations import override
from eprllib import logger

# TODO: Add here the availability to run EnergyPlus in Parallel.
# Run EnergyPlus in Parallel
# EnergyPlus is a multi-thread application but is not optimized for parallel runs on multi-core 
# machines. However, multiple parametric runs can be launched in parallel if these runs are 
# independent. One way to do it is to create multiple folders and copy files EnergyPlus.exe, 
# Energy+.idd, DElight2.dll, libexpat.dll, bcvtb.dll, EPMacro.exe (if macros are used), 
# ExpandObjects.exe (if HVACTemplate or Compact HVAC objects are used) from the original EnergyPlus 
# installed folder. Inside each folder, copy IDF file as in.idf and EPW file as in.epw, then run 
# EnergyPlus.exe from each folder. This is better handled with a batch file. If the Energy+.ini file 
# exists in one of the created folders, make sure it is empty or points to the current EnergyPlus folder.

# EP-Launch now does this automatically for Group or Parametric-Preprocessor runs.

# The benefit of run time savings depends on computer configurations including number of CPUs, 
# CPU clock speed, amount of RAM and cache, and hard drive speed. To be time efficient, the number 
# of parallel EnergyPlus runs should not be more than the number of CPUs on a computer. The EnergyPlus 
# utility program EP-Launch is being modified to add the parallel capability for group simulations. The 
# long term goal is to run EnergyPlus in parallel on multi-core computers even for a single EnergyPlus 
# run without degradation to accuracy.

class Environment(MultiAgentEnv):
    """
    The BaseEnvironment class represents a multi-agent environment for 
    reinforcement learning tasks related to building energy simulation using 
    EnergyPlus software. It inherits from the MultiAgentEnv class, which 
    suggests that it supports multiple agents interacting with the environment.

    Attributes:
        env_config (Dict[str, Any]): Configuration settings for the environment, 
            such as the list of agent IDs, action spaces, observable variables, 
            actuators, meters, and other EnergyPlus-related settings.
        episode_counter (int): Counter for the number of episodes.
        queues (Dict[str, Queue]): Queues for communication between the environment 
            and EnergyPlus.
        runner (BaseRunner): Instance of the EnergyPlusRunner class that handles 
            the EnergyPlus simulation.

    Methods:
        reset() -> Dict[str, Any]:
            Sets up a new episode of the environment. Increments the episode counter, 
            initializes queues for communication, and starts an instance of the 
            EnergyPlusRunner class.
        
        step(actions: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, float], Dict[str, bool], Dict[str, bool], Dict[str, Any]]:
            The core method where agents take actions, and the environment updates 
            its state accordingly. Processes the provided actions, communicates with 
            the EnergyPlus simulation, retrieves observations and information, 
            calculates rewards, and determines if the episode should terminate or truncate.
        
        close() -> None:
            Stops the EnergyPlus simulation when the environment is no longer needed.
        
        render() -> None:
            Placeholder method for rendering functionality.
    """
    def __init__(
        self,
        env_config: Dict[str, Any]
        ) -> None:
        """
        Initializes the BaseEnvironment class.

        Args:
            env_config (Dict[str, Any]): Configuration settings for the environment.
        """
        self.env_config = env_config
        
        # Episode and multiagent functions
        self.episode_fn: BaseEpisode = self.env_config['episode_fn'](self.env_config["episode_fn_config"])
        logger.debug(f"Episode configuration: {self.episode_fn.get_episode_config(self.env_config)}")
        self.connector_fn: BaseConnector = self.env_config['connector_fn'](self.env_config["connector_fn_config"])
        logger.debug(f"Connector configuration: {self.connector_fn.connector_fn_config}")
        
        # === AGENTS === #
        # Define all agent IDs that might even show up in your episodes.
        self.possible_agents = [key for key in self.env_config["agents_config"].keys()]
        logger.info(f"Possible agents: {self.possible_agents}")
        # If your agents never change throughout the episode, set
        # `self.agents` to the same list as `self.possible_agents`.
        self.agents = self.possible_agents
        # Otherwise, you will have to adjust `self.agents` in `reset()` and `step()` to whatever the
        # currently "alive" agents are.
        
        # asigning the configuration of the environment.
        self.reward_fn: Dict[str, Optional[BaseReward]] = {agent: None for agent in self.agents}
        
        self.trigger_fn: Dict[str, BaseTrigger] = {
            agent: self.env_config["agents_config"][agent]["trigger"]['trigger_fn'](
                self.env_config["agents_config"][agent]["trigger"]["trigger_fn_config"]
            )
            for agent in self.agents
        }
        self.filter_fn: Dict[str, BaseFilter] = {
            agent: self.env_config["agents_config"][agent]["filter"]['filter_fn'](
                self.env_config["agents_config"][agent]["filter"]["filter_fn_config"]
            )
            for agent in self.agents
        }
        
        # asignation of environment action space.
        action_space: Dict[str, Any] = {agent: None for agent in self.agents}
        for agent in self.agents:
            
            assert self.trigger_fn[agent] is not None, f"Trigger function for agent {agent} is not defined."
            
            action_space[agent] = self.trigger_fn[agent].get_action_space_dim()
            
        self.action_space = spaces.Dict(action_space)
        logger.debug(f"Action space: {self.action_space}")
        
        
        # Buffer to save the last "history_len" timesteps of observations
        self.history_len: Dict[str, int] = {}
        for agent in self.agents:
            self.history_len[agent] = self.env_config["agents_config"][agent]["observation"].get("history_len", 1) # Default history length is 1
        
            assert self.history_len[agent] > 0, f"History length must be greater than 0. Agent: {agent}, History length: {self.history_len[agent]}"
            assert isinstance(self.history_len[agent], int), f"History length must be an integer. Agent: {agent}, History length: {self.history_len[agent]}"
        # Create a deque for each agent to store the last observations.
        # The deque will have a maximum length of `history_len`.
        self.observation_buffers: Dict[str, collections.deque[Any]] = {
            agent:collections.deque(maxlen=self.history_len[agent] ) for agent in self.agents
        }
        # asignation of environment observation space.
        observation_space: Dict[str, Any]|spaces.Dict = {}
        original_obs_space: spaces.Dict = self.connector_fn.get_all_agents_obs_spaces_dict(self.env_config)
        logger.info(f"Originar observation space: {original_obs_space}")
        
        for agent_id, obs_space_per_agent in original_obs_space.items():
            logger.info(f"Observation space for agent {agent_id}: {obs_space_per_agent}")
            logger.info(f"History length for agent {agent_id}: {self.history_len[agent_id]}")
            if self.history_len[agent_id] > 1:
                # Asumiendo que obs_space_per_agent es un Box(shape=(feature_dim,))
                # Si es un Box(shape=(X, Y)), necesitarás aplanar o ajustar.
                assert isinstance(obs_space_per_agent, spaces.Box), f"Observation space for agent {agent_id} must be a Box space."
                assert obs_space_per_agent.shape is not None, f"Observation space for agent {agent_id} must have a defined shape."
                
                feature_dim: int = obs_space_per_agent.shape[0] # La dimensión de una única observación
                observation_space[agent_id] = spaces.Box(
                    low=obs_space_per_agent.low.min(), # Ajusta si tus rangos son por característica
                    high=obs_space_per_agent.high.max(), # Ajusta si tus rangos son por característica
                    shape=(self.history_len[agent_id], feature_dim), # Nueva forma: (N, feature_dim)
                    dtype=obs_space_per_agent.dtype
                )
                logger.info(f"Observation space for agente {agent_id} with history: {observation_space[agent_id]}")
            else:
                observation_space[agent_id] = obs_space_per_agent
                logger.info(f"Observation space for agent {agent_id} without history: {observation_space[agent_id]}")
        
        self.observation_space = spaces.Dict(observation_space)
        logger.info(f"Observation space: {self.observation_space}")
        
        # super init of the base class (after the previos definition to avoid errors with agents argument).
        super().__init__()
        
        # EnergyPlusRunner class and queues for communication between MDP and EnergyPlus.
        self.runner: Optional[EnvironmentRunner] = None
        self.obs_queue: Optional[Queue[Dict[str, Any]]] = None
        self.act_queue: Optional[Queue[Any]] = None
        self.infos_queue: Optional[Queue[Dict[str, Any]]] = None
        
        # === CONTROLS === #
        # variable for the registry of the episode number.
        self.episode = -1
        self.timestep = 0
        self.terminateds = False
        self.truncateds = False
        self.env_config['num_time_steps_in_hour'] = 0
        # dict to save the last observation and infos in the environment.
        self.last_obs: Dict[str, Any] = {agent: [] for agent in self.agents}
        self.last_infos: Dict[str, Any] = {agent: {} for agent in self.agents}
        
        # === DATA MANAGEMENT === #
        # output_path: Path to save the EnergyPlus simulation results.
        self.output_path = self.env_config["output_path"]
        if self.output_path is None:
            self.output_path = tempfile.gettempdir()
        logger.info(f"Output path for EnergyPlus simulation results: {self.output_path}")
        
        # If epjson_path is a IDF file, transform it into a epJSON file.
        if env_config['epjson_path'].endswith(".idf"):
            print("WARNING: Consider using epJSON files for full compatibility with the Episodes API.")

        logger.info("Environment initialization complete")
        
    @override(MultiAgentEnv)
    def reset(
        self, *,
        seed: Optional[int] = None,
        options: Optional[Dict[str, Any]] = None
        ):
        """
        Sets up a new episode of the environment.

        Returns:
            Dict[str, Any]: Initial observations for each agent.
        """
        # Call super's `reset()` method to (maybe) set the given `seed`.
        super().reset(seed=seed, options=options)
        # Set a list for the agents that are currently in the environment.
        # This list will be used to initialize the reward parameters.
        self.agents_to_inizialize_reward_parameters = self.possible_agents.copy()
        # Increment the counting of episodes in 1.
        self.episode += 1
        # saving the episode in the env_config to use across functions.
        self.env_config['episode'] = self.episode
        # stablish the timestep counting in zero.
        self.timestep = 0
        # Condition of truncated episode
        if not self.truncateds:
            # Condition implemented to restart a new epsiode when simulation is completed and 
            # EnergyPlus Runner is already inicialized.
            if self.runner is not None:
                self.close()
            # Define the queues for flow control between MDP and EnergyPlus threads in a max size 
            # of 1 because EnergyPlus timestep will be processed at a time.
            self.obs_queue = Queue(maxsize=1)
            self.act_queue = Queue(maxsize=1)
            self.infos_queue = Queue(maxsize=1)
            
            # episode_config_fn: Function that take the env_config as argument and upgrade the value
            # of env_config['epjson'] (str). Buid-in function allocated in tools.ep_episode_config
            try:
                self.env_config = self.episode_fn.get_episode_config(self.env_config)
                self.agents = self.episode_fn.get_episode_agents(self.env_config, self.possible_agents)
                logger.debug(f"Episode configuration: {self.env_config}")
            except (ValueError, FileNotFoundError):
                raise ValueError("The episode configuration is not valid. Please check the episode configuration.")
            
            # Update the rewards functions. The parameters of the reward function could be modify by the episode function, due that it's update here.
            for agent in self.agents:
                self.reward_fn[agent] = self.env_config["agents_config"][agent]["reward"]['reward_fn'](self.env_config["agents_config"][agent]["reward"]["reward_fn_config"])
            
            # Start EnergyPlusRunner whith the following configuration.
            self.runner = EnvironmentRunner(
                episode = self.episode,
                env_config = self.env_config,
                obs_queue = self.obs_queue,
                act_queue = self.act_queue,
                infos_queue = self.infos_queue,
                agents = self.agents,
                filter_fn = self.filter_fn,
                trigger_fn = self.trigger_fn,
                connector_fn = self.connector_fn
            )
            # Divide the thread in two in this point.
            self.runner.start()
            # Wait untill an observation and an infos are made, and get the values.
            self.runner.obs_event.wait()
            self.last_obs = self.obs_queue.get()
            self.runner.infos_event.wait()
            self.last_infos = self.infos_queue.get()
            
            # If the history length is greater than 1, we need to create a buffer for each agent
            # to store the last observations.
            for agent in self.agents:
                if self.history_len[agent] > 1:
                    self.observation_buffers[agent].clear()
                    # If the last observation is not empty, we append it to the buffer.
                    # This is done to ensure that the buffer has the last `history_len` observations.
                    single_obs_array = np.array(self.last_obs[agent])
                    for _ in range(self.history_len[agent]):
                        self.observation_buffers[agent].append(single_obs_array)
                    self.last_obs[agent] = np.array(self.observation_buffers[agent])
        # If the history length is greater than 1, we need to return the last `history_len` observations
        # for each agent.
        obs = self.last_obs
        infos = self.last_infos
        
        # set initial parameters for the reward function.
        for agent in obs.keys():
            reward_fn_instance = self.reward_fn.get(agent)
            if reward_fn_instance is not None:
                # set the initial parameters for the reward function.
                assert isinstance(reward_fn_instance, BaseReward), f"Reward function for agent {agent} must be a BaseReward instance."
                
                reward_fn_instance.set_initial_parameters(
                    agent,
                    self.connector_fn.obs_indexed[agent]
                    )
                logger.debug(f"Reward function for agent {agent} initialized with infos: {infos[agent]}")
                self.agents_to_inizialize_reward_parameters.remove(agent)
                logger.debug(f"Reward function initial parameters for agent {agent} initialized.")
                logger.debug(f"The agents to inizialize reward parameters: {self.agents_to_inizialize_reward_parameters}")
            else:
                logger.warning(f"No reward function defined for agent {agent}.")
                pass
        
        self.terminateds = False
        self.truncateds = False
            
        return obs, infos

    @override(MultiAgentEnv)
    def step(
        self, 
        actions: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, float], Dict[str, bool], Dict[str, bool], Dict[str, Any]]:
        """
        The core method where agents take actions, and the environment updates its state accordingly.

        Args:
            actions (Dict[str, Any]): Actions taken by each agent.

        Returns:
            Tuple[Dict[str, Any], Dict[str, float], Dict[str, bool], Dict[str, bool], Dict[str, Any]]: 
                Observations, rewards, done flags, truncated flags, and additional info for each agent.
        """
        # ===CONTROLS=== #
        # terminated variable is used to determine the end of a episode. Is stablished as False until the
        # environment present a terminal state.
        terminated: Dict[str, bool] = {}
        truncated: Dict[str, bool] = {}
        # Truncate the simulation RunPeriod into shorter episodes defined in days. Default: 0
        cut_episode_len: int = self.env_config['cut_episode_len']
        if self.env_config["evaluation"]:
            self.truncateds = False
        elif cut_episode_len == 0:
            self.truncateds = False
        else:
            # cut_episode_len_timesteps = cut_episode_len * 24 * self.env_config['num_time_steps_in_hour']
            if self.timestep % cut_episode_len == 0:
                self.truncateds = True
                logger.debug(f"Episode truncated after {cut_episode_len} timesteps.")
        # timeout is set to 10s to handle the time of calculation of EnergyPlus simulation.
        # timeout value can be increased if EnergyPlus timestep takes longer.
        timeout = self.env_config["timeout"]
        
        # check for simulation errors.
        assert self.runner is not None, "EnergyPlus Runner is not initialized. Please call reset() first."
        if self.runner.failed():
            self.terminateds = True
            logger.error("EnergyPlus failed with an error!")
        # simulation_complete is likely to happen after last env step()
        # is called, hence leading to waiting on queue for a timeout.
        if self.runner.simulation_complete:
            # if the simulation is complete, the episode is ended.
            self.terminateds = True
            # we use the last observation as a observation for the timestep.
            obs = self.last_obs
            infos = self.last_infos
            logger.info(f"Simulation completed after {self.timestep} timesteps.")

        # if the simulation is not complete, enqueue action (received by EnergyPlus through 
        # dedicated callback) and then wait to get next observation.
        else:
            try:
                assert isinstance(self.act_queue, Queue), "Action queue is not initialized. Please call reset() first."
                assert isinstance(self.obs_queue, Queue), "Observation queue is not initialized. Please call reset() first."
                assert isinstance(self.infos_queue, Queue), "Infos queue is not initialized. Please call reset() first."
                # Send the action to the EnergyPlus Runner flow.
                self.act_queue.put(actions)
                self.runner.act_event.set()
                # Modify the quantity of agents in the environment for this timestep, if apply.
                self.runner.agents = self.episode_fn.get_timestep_agents(self.env_config, self.possible_agents)
                
                # Get the return observation and infos after the action is applied.
                self.runner.obs_event.wait(timeout=timeout)
                current_obs = self.obs_queue.get(timeout=timeout)
                self.runner.infos_event.wait(timeout=timeout)
                infos = self.infos_queue.get(timeout=timeout)
                
                # If the history length is greater than 1, we need to create a buffer for each agent
                # to store the last observations.
                obs = {agent: np.array([]) for agent in self.agents}
                for agent in self.agents:
                    if self.history_len[agent] > 1:
                        # Append the observation to the buffer.
                        self.observation_buffers[agent].append(np.array(current_obs[agent]))
                        # Create the observation dict with the last `history_len` observations.
                        obs[agent] = np.array(self.observation_buffers[agent])
                    else:
                        obs[agent] = current_obs[agent]
                # logger.info(f"Observation for normal timestep {self.timestep}")
                # logger.info(f"Infos for timestep {self.timestep}: {infos}")
                # Check if the agents to initialize reward parameters is not empty.
                # If it is not empty, we initialize the reward parameters for each agent.
                if self.agents_to_inizialize_reward_parameters is not Empty:
                    for agent in obs.keys():
                        if agent in self.agents_to_inizialize_reward_parameters:
                            reward_fn_instance = self.reward_fn.get(agent)
                            if reward_fn_instance is not None:
                                reward_fn_instance.set_initial_parameters(
                                    agent,
                                    self.connector_fn.obs_indexed[agent]
                                    )
                                self.agents_to_inizialize_reward_parameters.remove(agent)
                
            except (Full, Empty):
                # logger.info("Queue timeout, ending episode early.")
                # Set the terminated variable into True to finish the episode.
                self.terminateds = True
                # We use the last observation as a observation for the timestep.
                obs = self.last_obs
                infos = self.last_infos
                # logger.info(f"Observation for terminal timestep {self.timestep}")
        
        # Calculate the reward in the timestep
        reward_dict: Dict[str, float] = {}
        for agent in obs.keys():
            # The reward function instance can be None if not defined for an agent.
            # We use .get() for safer access and check for None.
            
            assert type(agent) == str, f"Agent {agent} is not a string."
            
            reward_fn_instance = self.reward_fn.get(agent)
            if reward_fn_instance:
                if self.history_len[agent] > 1:
                    # If the history length is greater than 1, we use only the last observation in the buffer.
                    obs_to_reward = np.array(self.last_obs[agent][-1])
                    next_obs_to_reward = np.array(self.observation_buffers[agent][-1])
                else:
                    obs_to_reward = np.array(self.last_obs[agent])
                    next_obs_to_reward = np.array(obs[agent])
                reward_dict[agent] = reward_fn_instance.get_reward(
                    obs_to_reward,
                    actions.get(agent),
                    next_obs_to_reward,
                    self.terminateds,
                    self.truncateds
                )
            else:
                # Assign a default reward of 0.0 if no reward function is found.
                reward_dict[agent] = 0.0
        
        terminated["__all__"] = self.terminateds
        truncated["__all__"] = self.truncateds
        
        # Upgrade last observation and infos dicts.
        self.last_obs = obs
        self.last_infos = infos
        
        # increment the timestep in 1.
        self.timestep += 1
        
        return obs, reward_dict, terminated, truncated, infos

    @override(MultiAgentEnv)
    def close(self):
        """
        Stops the EnergyPlus simulation when the environment is no longer needed.
        """
        if self.runner is not None:
            self.runner.stop()
    
    @override(MultiAgentEnv)
    def render(self, mode:str="human"):
        """
        Placeholder method for rendering functionality.
        """
        pass
    
    @staticmethod
    def from_checkpoint(
        cls,
        path: str
    ) -> "Environment":
        msg = "This method is not implemented yet. Please implement it in the child class."
        logger.error(msg)
        raise NotImplementedError(msg)
    
    def get_default_config(cls) -> EnvironmentConfig:
        return EnvironmentConfig()


"""
Reward functions in eprllib
============================

The reward functions are called in the ``step()`` method in the environment iteratively for
each agent present in the respective timestep.

This module contains the reward functions used in reinforcement learning and 
implemented in eprllib.

You can implement your own reward function by creating a new class that inherits from
:class:`~eprllib.Agents.Rewards.BaseReward` and overriding the ``get_reward`` method.

The module includes the following classes:

- :class:`~eprllib.Agents.Rewards.BaseReward`: The base class for creating reward functions.
- :class:`~eprllib.Agents.Rewards.ASHRAE55SimpleModel`: Contains classes to calculate rewards based on ASHRAE-55 comfort model.
- :class:`~eprllib.Agents.Rewards.CEN15251`: Contains classes to calculate rewards based on CEN-15251 comfort model.
- :class:`~eprllib.Agents.Rewards.EnergyRewards`: Contains classes to calculate rewards based on energy.

Also, it is possible to combine the previous rewards methods for multi-objective reward funtions that
are normaly used in building control (e.g. use comfort and energy optimization). Implementations in 
eprllib are:

- :class:`~eprllib.Agents.Rewards.EnergyAndAshrae55SimpleModel`
- :class:`~eprllib.Agents.Rewards.EnergyAndCEN15251`

These classes are used to provide feedback to the agents, guiding them to optimize their policies based on 
the defined reward criteria.
"""



"""
Specification for the trigger functions
===========================================================
This module defines the `TriggerSpec` class, which is used to specify the configuration of trigger 
functions for agents in reinforcement learning environments.
"""
from typing import Dict, Any, Type, Optional
from eprllib.Agents.Triggers.BaseTrigger import BaseTrigger
from eprllib import logger

class TriggerSpec:
    """
    TriggerSpec is the base class for a trigger specification to safe configuration of the object.
    """
    def __init__(
        self,
        trigger_fn: Optional[Type[BaseTrigger]] = None,
        trigger_fn_config: Dict[str, Any] = {},
    ):
        """
        Construction method.
        
        Args:
            trigger_fn (BaseTrigger): The trigger function takes the arguments agent_id, observation and returns the
            observation filtered. See ``eprllib.Agents.Triggers`` for examples.
            
            trigger_fn_config (Dict[str, Any]): The configuration of the trigger function.
            
        Raises:
            ValueError: If the trigger function is not a subclass of BaseTrigger or if the configuration is not a dictionary.
            NotImplementedError: If the trigger function is NotImplemented.
            TypeError: If the trigger function is not a subclass of BaseTrigger.
        """
        self.trigger_fn = trigger_fn
        self.trigger_fn_config = trigger_fn_config
    
    def __getitem__(self, key:str) -> Any:
        return getattr(self, key)

    def __setitem__(self, key:str, value:Any) -> None:
        valid_keys = self.__dict__.keys()
        if key not in valid_keys:
            msg = f"Invalid key: {key}."
            logger.error(msg)
            raise KeyError(msg)
        setattr(self, key, value)
    
    def build(self) -> Dict[str, Any]:
        """
        This method is used to build the TriggerSpec object.
        """
        if self.trigger_fn == None:
            msg = "The trigger function must be defined."
            logger.error(msg)
            raise ValueError(msg)
            
        return vars(self)



"""
Centralized Agents Connector
=============================

A central agent takes the observations of all the agents involved in the environment and concatenates them
to create a single observation. After transforming the multiple observations into one, this is used in the 
central policy to select multiple discrete actions, one for each agent.

To avoid parameter repetitions in the central agent observation, only implement an observation parameter
in a single agent. For example, if two agents are present in the same thermal zone and both of them have
access to the thermal zone mean air temperature, only declare this parameter in one of them.
"""
from gymnasium.spaces import Box
import numpy as np
from typing import Any, Dict, Tuple
from eprllib.AgentsConnectors.BaseConnector import BaseConnector
from eprllib.Utils.annotations import override
from eprllib.Utils.connector_utils import (
    set_variables_in_obs,
    set_internal_variables_in_obs,
    set_meters_in_obs,
    set_simulation_parameters_in_obs,
    set_zone_simulation_parameters_in_obs,
    set_prediction_variables_in_obs,
    set_other_obs_in_obs,
    set_actuators_in_obs,
    set_user_occupation_forecast_in_obs
    )
from eprllib import logger

class CentralizedConnector(BaseConnector):
    def __init__(
        self,
        connector_fn_config: Dict[str,Any]
        ):
        """
        This class implements a centralized policy for the observation function.

        Args:
            connector_fn_config (Dict[str, Any]): The configuration dictionary for the observation function.
            This must contain the key 'number_of_agents_total', which represents the maximum
            quantity to which the policy is prepared. It is related to the unitary vector.
        """
        super().__init__(connector_fn_config)
    
    @override(BaseConnector)
    def get_agent_obs_dim(
        self,
        env_config: Dict[str, Any],
        agent: str
    ) -> Box:
        """
        Construct the observation space of the environment.

        Args:
            env_config (Dict[str, Any]): The environment configuration dictionary.
            agent (str, optional): The agent identifier.

        Returns:
            gym.Space: The observation space of the environment.
        """
        obs_space_len: int = 0
        self.obs_indexed[agent] = {}
        
        self.obs_indexed[agent], obs_space_len = set_variables_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_internal_variables_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_meters_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_simulation_parameters_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_zone_simulation_parameters_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_prediction_variables_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_other_obs_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_actuators_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_user_occupation_forecast_in_obs(env_config, agent, self.obs_indexed[agent], obs_space_len)
        
        assert obs_space_len > 0, "The observation space length must be greater than 0."
        assert len(self.obs_indexed[agent]) == obs_space_len, "The observation space length must be equal to the number of indexed observations."
        # obs_space_len += 1
        logger.debug(f"Observation space length for agent {agent}: {obs_space_len}")
        
        return Box(float("-inf"), float("inf"), (obs_space_len, ))
    
    @override(BaseConnector)
    def get_agent_obs_indexed(
        self,
        env_config: Dict[str, Any],
        agent: str
    ) -> Dict[str, int]:
        """
        Get a dictionary of the agent observation parameters and their respective index in the observation array.

        :param env_config: Environment configuration.
        :type env_config: Dict[str, Any]
        :param agent: Agent identifier, optional.
        :type agent: str, optional
        :return: Agent observation spaces.
        :rtype: gym.spaces.Space
        """
        if self.obs_indexed == {}:
            self.get_agent_obs_dim(env_config, agent)
        return self.obs_indexed[agent]
    
    @override(BaseConnector)
    def set_top_level_obs(
        self,
        env_config: Dict[str, Any],
        agent_states: Dict[str,Dict[str,Any]],
        dict_agents_obs: Dict[str,Any],
        infos: Dict[str, Dict[str, Any]],
        is_last_timestep: bool = False
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]:
        
        # agents in this timestep
        agent_list = [key for key in dict_agents_obs.keys()]
        # Add agent indicator for the observation for each agent
        agents_obs = {"central_agent": np.array([], dtype='float32')}
        
        for agent in agent_list:
            agents_obs["central_agent"] = np.concatenate(
                (
                    agents_obs["central_agent"],
                    dict_agents_obs[agent]
                ),
                dtype='float32'
            )
            
        return agents_obs, infos, True



"""
Windows Opening Triggers
========================

This module contains classes to implement window opening triggers for controlling actuators in the environment.
"""
import gymnasium as gym
import numpy as np
from typing import Any, Dict, List, Tuple
from eprllib.Agents.Triggers.BaseTrigger import BaseTrigger
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override
from eprllib.Utils.agent_utils import get_agent_name, config_validation
from eprllib import logger

class WindowsOpeningDiscreteTrigger(BaseTrigger):
    REQUIRED_KEYS: Dict[str, Any] = {
        "window_actuator": Tuple[str, str, str],
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any]
    ):
        """
        This class implements the window opening action function.

        Args:
            trigger_fn_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - window_actuator (Tuple[str, str, str]): The configuration for the window actuator.
        """
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
        
        self.agent_name = None
        self.window_actuator = None
    
    @override(BaseTrigger)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Discrete(11)
    
    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str, Any]:
        """
        Transform the agent action to actuator action.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        if self.agent_name is None:
            self.agent_name = get_agent_name(actuators)
            self.window_actuator = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['window_actuator'][0],
                self.trigger_fn_config['window_actuator'][1],
                self.trigger_fn_config['window_actuator'][2]
            )
        
        assert self.window_actuator is not None, "Window actuator name has not been initialized."
        
        actuator_dict_actions = {actuator: None for actuator in actuators}
        actuator_dict_actions.update({self.window_actuator: action / 10})
        
        # Check if there is an actuator_dict_actions value equal to None.
        for actuator in actuator_dict_actions:
            if actuator_dict_actions[actuator] is None:
                msg = f"The actuator {actuator} is not in the list of actuators: \n{actuators}.\nThe actuator dict is: \n{actuator_dict_actions}"
                logger.error(msg)
                raise ValueError(msg)
            
        return actuator_dict_actions
    
    @override(BaseTrigger)
    def get_actuator_action(self, action: float | int, actuator: str) -> Any:
        """
        Get the actuator action.

        Args:
            action (float | int): The action to be applied.
            actuator (str): The actuator to be controlled.

        Returns:
            Any: The action for the actuator.
        """
        return action

    @override(BaseTrigger)
    def action_to_goal(self, action: int | float) -> int | float:
        """
        This method is used to transform the action to a goal. The goal is used to define the reward.

        Args:
            action (Any): The action to be transformed.

        Returns:
            Any: The transformed action.
        """
        return action
    

class WindowsOpeningContinousTrigger(BaseTrigger):
    REQUIRED_KEYS: Dict[str, Any] = {
        "window_actuator": Tuple[str, str, str],
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any]
    ):
        """
        This class implements the window opening action function.

        Args:
            trigger_fn_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - window_actuator (Tuple[str, str, str]): The configuration for the window actuator.
        """
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
        
        self.agent_name = None
        self.window_actuator = None
    
    @override(BaseTrigger)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32)
    
    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str, Any]:
        """
        Transform the agent action to actuator action.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        if self.agent_name is None:
            self.agent_name = get_agent_name(actuators)
            self.window_actuator = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['window_actuator'][0],
                self.trigger_fn_config['window_actuator'][1],
                self.trigger_fn_config['window_actuator'][2]
            )
        
        assert self.window_actuator is not None, "Window actuator name has not been initialized."
        
        actuator_dict_actions = {actuator: None for actuator in actuators}
        actuator_dict_actions.update({self.window_actuator: action})
        
        # Check if there is an actuator_dict_actions value equal to None.
        for actuator in actuator_dict_actions:
            if actuator_dict_actions[actuator] is None:
                msg = f"The actuator {actuator} is not in the list of actuators: \n{actuators}.\nThe actuator dict is: \n{actuator_dict_actions}"
                logger.error(msg)
                raise ValueError(msg)
            
        return actuator_dict_actions
    
    @override(BaseTrigger)
    def get_actuator_action(self, action: float | int, actuator: str) -> Any:
        """
        Get the actuator action.

        Args:
            action (float | int): The action to be applied.
            actuator (str): The actuator to be controlled.

        Returns:
            Any: The action for the actuator.
        """
        return action

    @override(BaseTrigger)
    def action_to_goal(self, action: int | float) -> int | float:
        """
        This method is used to transform the action to a goal. The goal is used to define the reward.

        Args:
            action (Any): The action to be transformed.

        Returns:
            Any: The transformed action.
        """
        return action


"""
Exhaust Fan triggers
=====================

This module contains classes to implement triggers for controlling exhaust fan actuators in the environment.
"""
import gymnasium as gym
import numpy as np
from typing import Any, Dict, List, Tuple, Optional
from eprllib.Agents.Triggers.BaseTrigger import BaseTrigger
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override
from eprllib.Utils.agent_utils import get_agent_name, config_validation
from eprllib import logger

class ExhaustFanTrigger(BaseTrigger):
    REQUIRED_KEYS: Dict[str, Any] = {
        "modes": List[float|int],
        "exhaust_fan_actuator": Tuple[str, str, str]
    }
    
    def __init__(
        self,
        trigger_fn_config: Dict[str, Any]
    ):
        """
        This class implements the Exhaust Fan actions.

        Args:
            trigger_fn_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - modes (List[float]): The flow factor to modify the maximum flow of the exhaust fan. The order 
                in the list corresponds with mode of the Fan, usually mode 0 is off and mode 1 has the lower 
                flow factor. No more than 11 modes are allowed.
                - exhaust_fan_actuator (Tuple[str, str, str]): The configuration for the exhaust fan actuator.
                The first element is the agent name, the second is the actuator name, and the third is the actuator type.
        Raises:
            ValueError: If the configuration is not valid, or if the modes are not in the range [0, 1], or if the length of the modes is larger than 11.
        """
        # Validate the config.
        config_validation(trigger_fn_config, self.REQUIRED_KEYS)
        
        super().__init__(trigger_fn_config)
        
        self.agent_name = None
        self.modes: List[float] = trigger_fn_config['modes']
        self.exhaust_fan_actuator: Optional[str] = None
        
        #  Check if the lenght of the modes are larger than 11 (that is the action space for this class).
        if len(self.modes) > 11:
            msg = f"The lenght of the modes must be less than 11. The lenght of the modes is {len(self.modes)}."
            logger.error(msg)
            raise ValueError(msg)
        
        # Check that all the elements in self.modes list are floats in the range [0, 1].
        for mode in self.modes:
            if mode < 0 or mode > 1:
                msg = f"The mode {mode} is not in the range [0, 1]."
                logger.error(msg)
                raise ValueError(msg)
    
    @override(BaseTrigger)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Discrete(11)
    
    @override(BaseTrigger)
    def agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
            
        Raises:
            ValueError: If the actuator is not in the list of actuators.
        """
        if self.agent_name is None:
            self.agent_name = get_agent_name(actuators)
            self.exhaust_fan_actuator = get_actuator_name(
                self.agent_name,
                self.trigger_fn_config['exhaust_fan_actuator'][0],
                self.trigger_fn_config['exhaust_fan_actuator'][1],
                self.trigger_fn_config['exhaust_fan_actuator'][2]
            )
            
        actuator_dict_actions: Dict[str, Any] = {actuator: None for actuator in actuators}
        
        # TODO: This can be not beneficial for the selection of an action, because introduce noise
        # in the effect of the actions. The best way of avoid this (I think) it to introduce a mask
        # for the actions that reduce the probability of choose the actions that are not in the modes.
        if action >= len(self.modes):
            action_list = [_ for _ in range(len(self.modes))]
            action = np.random.choice(action_list)
        
        # The actuator name is optional in the __init__ but it is initialized in the first call to this method.
        # We need to assert that it is not None to avoid a type error.
        assert self.exhaust_fan_actuator is not None, "Exhaust fan actuator name has not been initialized."

        actuator_dict_actions.update({
            self.exhaust_fan_actuator: self.modes[action],
        })

        # Check if there is an actuator_dict_actions value equal to None.
        for actuator in actuator_dict_actions:
            if actuator_dict_actions[actuator] is None:
                msg = f"The actuator {actuator} is not in the list of actuators: \n{actuators}.\nThe actuator dict is: \n{actuator_dict_actions}"
                logger.error(msg)
                raise ValueError(msg)
        
        return actuator_dict_actions
    
    @override(BaseTrigger)
    def get_actuator_action(self, action:float|int, actuator: str) -> Any:
        """
        This method is used to get the actions of the actuators after transform the 
        agent action to actuator action.

        Args:
            action (float|int): The action of the agent.
            actuator (str): The name of the actuator.

        Returns:
            Any: The action of the actuator.
        """
        return action



"""
Agents connectors
==================

This module contains connectors for agents, which are responsible for combining and transforming 
agents' observations to provide flexible configurations for communication between agents. 

The module includes the following classes:

- BaseConnector: The base class for creating connector functions.
- DefaultConnector: Implements the default connector class that allows the combination of agents' observations.
- CentralizedConnector: Implements a centralized connector where a central agent takes the observations of all agents.
- FullySharedParametersConnector: Implements a fully shared parameters policy for the observation function.
- HierarchicalConnector: Implements a hierarchical connector with two levels of hierarchy.
- IndependentConnector: Implements the default observation function where each agent has its own observation space.

These connectors are used to provide different configurations for multi-agent environments, enabling various 
communication and observation strategies.
"""